/*************************************************************************/
/**************************** bfs.c **************************************/
/*************************************************************************/
#include <stdio.h>
#include "queue.h"
#include "part.h"

void enqueue(int, Q_TYPPTR);
int dequeue(Q_TYPPTR);

extern int num;
extern Q_TYPPTR get_q();
extern struct adj_node *adj_list[];

int dmax = 0;				/* maximum distance in bfs      */
int distance[MAX_VERTS];		/* distance from search root 	*/
					/* if distance[i]==0            */
					/* then i is not adj to z       */ 
int marked[MAX_VERTS];		/* boolean - marked or not?  	*/
int pred[MAX_VERTS];			/* predecessor tree		*/


void bfs(int lmng, int *vertices)
/*  lmng;*/				/* rank of last meaningful	*/
/*  *vertices;*/			/* vertex nums (indexed by rank)*/
{
	int i, c, node;
	int rank[MAX_VERTS];		/* Contains the rank of vertex  */
					/* i in the ordering 'vertices' */
	struct adj_node *atemp;
	Q_TYPPTR q;

	for (i=0; i<num; i++) {
		rank[vertices[i]] = i;
		marked[i] = FALSE;
		distance[i] = 0; 
		pred[i] = EMPTY;
	}
	q = get_q();
	marked[vertices[lmng]] = TRUE;
	enqueue(vertices[lmng], q);
	while ((c=dequeue(q)) != EMPTY) {
		for (atemp = adj_list[c]; atemp; atemp = atemp->next) {
			node = atemp->node_no;
			if (rank[node] > lmng)
				if (!marked[node]) {
					marked[node] = TRUE;
					distance[node] = distance[c] + 1;
					if (distance[node] > dmax)
						dmax = distance[node];
					pred[node] = c;
					enqueue(node, q);
				}
		}
	}
}

void print_distances(int lmng, int *vertices)
/*  lmng;*/				/* rank of last meaningful	*/
/*  *vertices;*/			/* vertex nums (indexed by rank)*/
{
	int i, j;
	printf("****************************************************\n");
	printf("*** Distances ******** (vertex numbers) ************\n");
	printf("****************************************************\n");
	printf("0: %d\n", vertices[lmng]); 
	for (j=1; j<=dmax; j++) {
	printf("****************************************************\n");
		printf("**%d: ", j); 
		for (i=lmng+1; i<num; i++)
			if (distance[vertices[i]] == j) 
				printf("%d ", vertices[i]);
		printf("\n");
	}
}
	


/*Q_TYPPTR vertex_q;
main()
{
	int i,j,k, num;
	char op[2], str[80];

	vertex_q = get_q();

	while (op[0] != 'q') {
		fgets(str, 80, stdin);
		sscanf(str, "%s %d %d %d\n", op, &k, &i, &j);
		switch (op[0]) {
			case 'e' : enqueue(k, vertex_q);
				   break;
			case 'd' : i = dequeue(vertex_q);
			           printf("dequeue: %d\n",i);
				   break;
			case 'n' : vertex_q = get_q();
				   break;
			case 'f' : free_queue(vertex_q);
				   break;
		}
		print_queue(vertex_q);
	}
}
*/
