#include <stdio.h>
#include <math.h>
#include "part.h"
#include "data_sets.h"

extern int NG_Count[];			/* [i]:  #ng up to placement i    */
extern int Tot_Cost[];			/* [i]:  #cut up to placement i   */
extern int Size_A[];			/* [i]:  # in A up to placement i */
extern int Size_B[];			/* [i]:  # in B up to placement i */

DATA_SET data_sets[NUM_DATA_SETS];

void init_data_sets()
{
	int i, j;

	for (i=0; i<NUM_DATA_SETS; i++) {
		data_sets[i].graph_count = 0;
		for (j=0; j<MAX_VERTICES; j++) {
			data_sets[i].total[j] = 0.0;
			data_sets[i].sqr_total[j] = 0.0;
		}
	}
}

void read_data_sets()
{
        extern int Regular_Graphs;
        extern int Reg_Degree;
        extern int Reg_Degree2;
        extern int Unbalanced_Regular_Graphs;
        extern int Unbalanced_Random_Graphs;
        extern int Geometric_Graphs;
        extern int num;
	extern double prob, prob2;
	extern FILE *outfile;

	int i, j, k;
	char fnames[NUM_DATA_SETS][80], pr[30], *home, subdir[30];
	char ds_names[NUM_DATA_SETS][30];
	FILE **fp;

	data_sets[NG_COUNT].data = NG_Count;
	strcpy(ds_names[NG_COUNT], "ng_count"); 
	data_sets[TOT_COST].data = Tot_Cost;
	strcpy(ds_names[TOT_COST], "tot_cost"); 
	data_sets[SIZE_A].data = Size_A;
	strcpy(ds_names[SIZE_A], "size_a"); 
	data_sets[SIZE_B].data = Size_B;
	strcpy(ds_names[SIZE_B], "size_b"); 

        home=(char *)malloc(80);
        /*******************************************************************/
        /** for this code to work, V must be a multiple of 100            **/
        /*******************************************************************/
        if (Regular_Graphs) {
                sprintf(pr, "%d", Reg_Degree);
        } else if (Unbalanced_Regular_Graphs) {
                sprintf(pr, "%d.%d", Reg_Degree, Reg_Degree2);
        } else if (Unbalanced_Random_Graphs) {
                sprintf(pr, "%.3f.%.3f", prob, prob2);
        } else if (Geometric_Graphs) {
                sprintf(pr, "%.3f.G", prob);
        } else
                sprintf(pr, "%.3f", prob);
        home = getenv("HOME");
        strcpy(subdir, "/greedy/density_functions/");
        /*strcpy(subdir, "/research/test0793/");*/

	for (i=0; i<NUM_DATA_SETS; i++) {
            sprintf(fnames[i],"%s%s%s.%d.%s",home,subdir,ds_names[i],num,pr);
	}

	for (i=0; i<NUM_DATA_SETS; i++) {
		fp = &data_sets[i].fp;
		if ((*fp = fopen(fnames[i], "r+")) != NULL) {
			fscanf(*fp," %d", &data_sets[i].graph_count);
			for (j=0; j<num; j++) {
				fscanf(*fp, "%d %lf %lf", &k, 
					&data_sets[i].total[j],
					&data_sets[i].sqr_total[j]);
			}
		} else if ((*fp = fopen(fnames[i], "w")) == NULL) 
			fprintf(outfile,"Error in %s file\n",fnames[i]);
		rewind(*fp);
	}
}

void update_data_sets()		/* called from nongreedy_stats() after */
{				/* each graph is completed.	       */
	extern int num;
	int i, j;

	for (i=0; i<NUM_DATA_SETS; i++) {
		data_sets[i].graph_count++;
		for (j=0; j<num; j++) {
			data_sets[i].total[j] += (double) data_sets[i].data[j];
			data_sets[i].sqr_total[j] += (double)
				(data_sets[i].data[j] * data_sets[i].data[j]);
		}
	}
}

void write_data_sets(int num_v)
{
	int i, j;
	double count, mean, var, std_dev;
	FILE *fp;

	for (i=0; i<NUM_DATA_SETS; i++) {
		fp = data_sets[i].fp;
		fprintf(fp," %d\n", data_sets[i].graph_count);
		for (j=0; j<num_v; j++) {
			fprintf(fp, "%d %.1f %.1f\n", j, 
				data_sets[i].total[j],
				data_sets[i].sqr_total[j]);
		}
		fprintf(fp, "MEAN/VARIANCE TABLE:\n");
		fprintf(fp, "-------------------:\n");
		count = (double) data_sets[i].graph_count;
		for (j=0; j<num_v; j++) {
			mean = data_sets[i].total[j] / count;
			var  =	(data_sets[i].sqr_total[j]/count) -
				(mean * mean);
			std_dev = sqrt(var);
			fprintf(fp, "%d %f %f %f\n", j, mean, var, std_dev);
		}
		fflush(fp);
		rewind(fp);
	}
}
