#include "part.h"

extern int num_processed;
extern int N_Interval;
extern int P_Interval;
extern int P_Prob;
extern int ng_cost[][MAX_VERTICES];
extern int node_cost[][MAX_VERTICES];
extern PG_DIST (*ng_distrib)[DISTRIB_SIZE];

double fixed_prob()
{
	return 0.05;
}

double find_prob(int node, int *dist_used)
{

	double ng_prob;
	int p_index;
	
	p_index = (int) (num_processed / P_Interval);

	ng_prob = ng_distrib[P_Prob][p_index].d[NUM_INTERVALS];
	*dist_used = NUM_INTERVALS;
	return ng_prob;
}

double find_prob_abs(int node, int *dist_used)
{

	double ng_prob = 0.0;
	int k, abs_diff, p_index;
	
	p_index = (int) (num_processed / P_Interval);
        abs_diff =    ABS(node_cost[SET_A][node] - node_cost[SET_B][node]);


	for (k=0; k<NUM_INTERVALS; k++) {
        	if (abs_diff == k) {
			ng_prob = ng_distrib[P_Prob][p_index].d[k];
			*dist_used = k;
			break;
		}
	}
	return ng_prob;
}


double find_prob_ng_10(int node, int *dist_used)
{

	extern int ng_count;
	double ng_prob = 0.0;
	int k, p_index;
	static int ng_in_first_10;
	
	p_index = (int) (num_processed / P_Interval);

	if (num_processed < 10) {
		ng_prob = ng_distrib[P_Prob][p_index].d[SIZE_INTERVAL_ARRAY-1];
		return ng_prob;
	} 
	if (num_processed == 10)
		ng_in_first_10 = ng_count;

	for (k=0; k<NUM_INTERVALS; k++) {
        	if (ng_in_first_10 == k) {
			ng_prob = ng_distrib[P_Prob][p_index].d[k];
			*dist_used = k;
			break;
		}
	}
	return ng_prob;
}

double find_prob_k(int node, int *dist_used)
{
	double ng_prob = 0.0;
	int k, ng_up, p_index;
	
	p_index = (int) (num_processed / P_Interval);
	ng_up =    ng_cost[SET_A][node]+ ng_cost[SET_B][node];

	for (k=0; k<NUM_INTERVALS; k++) {
        	if (ng_up == k) {
			ng_prob = ng_distrib[P_Prob][p_index].d[k];
			*dist_used = k;
			break;
		}
	}
	return ng_prob;
}

double find_prob_k_over_total(int node, int *dist_used)
{
	extern int Use_Standard_For_Middle;
	extern int N_Prob;
	extern PCT_NG_UP (*NG_Up_Results)[NG_UP_SIZE];

	double ng_prob = 0.0;
	int n_index, k, ng_up, total_up, p_index;
	double step, start, current_ng_pct;
	
	n_index = (int) (num_processed / N_Interval);
	p_index = (int) (num_processed / P_Interval);
	step = (NG_Up_Results[N_Prob][n_index].gr_pct -
		NG_Up_Results[N_Prob][n_index].ng_pct)
			/ (double) NUM_INTERVALS;
	start = NG_Up_Results[N_Prob][n_index].ng_pct;
	ng_up =    ng_cost[SET_A][node]+ ng_cost[SET_B][node];
	total_up = node_cost[SET_A][node]+ node_cost[SET_B][node];
	current_ng_pct = (total_up>0) ? (double) ng_up/(double) total_up
			: 0.0;

	if (!Use_Standard_For_Middle) {
		for (k=0; k<NUM_INTERVALS; k++) {
                       if (in_interval(current_ng_pct, start, step, k)) {
				ng_prob = ng_distrib[P_Prob][p_index].d[k];
				*dist_used = k;
				break;
			}
			if (k == (NUM_INTERVALS-1)) {
				ng_prob = ng_distrib[P_Prob][p_index].d[k];
				*dist_used = k;
			}
		}
	} else {
	     	if (in_interval(current_ng_pct, start, step, -1))
	    		ng_prob = ng_distrib[P_Prob][p_index].d[0];
	     	else if (in_interval(current_ng_pct, start, 
						     step, NUM_INTERVALS-1))
	    		ng_prob = 
			    ng_distrib[P_Prob][p_index].d[SIZE_INTERVAL_ARRAY-1];
	     	else
	    		ng_prob = 
			    ng_distrib[P_Prob][p_index].d[NUM_INTERVALS-1];
	}
	return ng_prob;
}
