#include <stdio.h>
#include <stdlib.h>

#define MAX_INT		2147483647
#define	KL		-7
#define	W		-8
#define	PG		-9
#define RAND 		-10
#define UNBAL_RAND 	-11
#define UNBAL_REG 	-12
#define REG		-13

extern long 	random();

/**********************************************************************/
/** This state information is used for initializing the random       **/
/** number generator.  (try "man 3 random" for more information)     **/
/**********************************************************************/
static long state1[32] = {
               3,
               0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
               0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
               0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
               0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
               0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
               0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
               0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
               0xf5ad9d0e, 0x8999220b, 0x27fb47b9
};

int	num_vertices;
int	num_graphs;
int	graph_type;
int	iterations;
int	iterations2;
int	reorder_type;		/* for kl postprocessing	*/
int	local_opt;		/* local opt. or not		*/
int	kl_opt;			/* aug w with kl 		*/
int	alg;			/* kl, pg, or w			*/
int	d_reg, d_reg2;
int     KL_PG_Flag;
long	seed;			/* for rand rum generator	*/
double	edge_prob;
double	edge_prob2;
double	time;			/* = 0.0 --> use iterations     */
				/* else run for <time> seconds  */

FILE 	*outfile;

/*****************************************************/
/********  initialize random number generator  *******/
/*****************************************************/
void randinit(unsigned cell)
{
	initstate(cell,state1,128);	/* 4/27/93 */
	/*srand48(cell);*/
}

void set_defaults()
{
	num_vertices = 500;
	graph_type= RAND;
	edge_prob = 0.5;
	edge_prob2 = 0.5;
	d_reg = 5;
	d_reg2 = 5;
	iterations = 5;
	iterations2 = 5;
	reorder_type = 7;
	local_opt = 1;
	kl_opt = 0;
	time = 0.0;
	num_graphs = 10;
	seed = 399;
	alg = PG;
	KL_PG_Flag = 0;
	outfile = stdout;
}

/*************************************************************************/
/** process_args:	the format is:					**/
/**		gi -n <num_v> -p <prob> -g <num_g> -s <seed>            **/
/**		  (-i <iterations [,iterations2]> || -t time)  -[kw(pg)]**/
/**		  -[r(ur)(ug)]						**/
/**		  -r <reorder_type> -l <local opt?> -f <file_name>	**/
/*************************************************************************/
void process_args(int argc, char *argv[])
{
	char *s;

	if(argc <2) {
	    printf(
		"usage: gi -n <num_v> \n"
		"          -p <prob1> (edge prob. or distance threshold)\n"
		"          -g <#graphs>\n"
		"          -d <d-reg>  (d-regular graphs)\n"
		"          -s <random seed> \n"
		"          -i <iterations> | -t <time> (choose one)"
	        "          -k | -w        (-k: random or line starts, -w: W"
				" Starts)\n"
	    	"          -[(ul): regular (ug): unbal_reg (ur): unbal rand] \n"
		"          -p2 <prob2> (for unbalanced random graphs\n)"
		"          -d2 <d2-reg-bipartite> (unbalanced regular)\n"
	    	"          -r <reorder_type> (irrelevant for heuristic runs)\n"
		"          -l <local opt?> (run simple local optz. after each)\n"
	    	" 	   -f <file_name>  (save input file as file_name)\n");
	    exit(1);
	}

	while(--argc > 0 && (*++argv)[0] == '-') {
	   for(s=argv[0]+1; *s != '\0'; s++)
		switch(*s) {
			case 'K' :  
				KL_PG_Flag = 1;
				break;
			case 'n' :  
				--argc;
				num_vertices = atoi((++argv)[0]);
				break;
			case 'p' :  
				if (*(s+1) == 'g') {
					alg = PG;
					if (num_vertices % 100 != 0) {
						printf("illegal num. of v.\n");
						exit(2);
					}
					s++;
				} else if (*(s+1) == '2') {
					--argc;
					edge_prob2 = atof((++argv)[0]);
					s++;
				} else {
					--argc;
					edge_prob = atof((++argv)[0]);
				}
				break;
			case 'd' :  
				if (*(s+1) == '2') {
					--argc;
					d_reg2 = atoi((++argv)[0]);
					s++;
				} else {
					--argc;
					d_reg = atoi((++argv)[0]);
				} 
				break;
			case 'u' :  
				if (*(s+1) == 'r') {
					graph_type = UNBAL_RAND;
					s++;
				} else  if (*(s+1) == 'g') {
					graph_type = UNBAL_REG;
					s++;
				} else  if (*(s+1) == 'l') {
					graph_type = REG;
					s++;
				} 
				break;
			case 'f' :  
				--argc;
				if ((outfile = fopen((++argv)[0],"w")) == NULL)
					printf("file error\n"), exit(1);
				break;
			case 'l' :
				--argc;
				if (*(s+1) == 'k')
					kl_opt = atoi((++argv)[0]);
				else
					local_opt = atoi((++argv)[0]);
				s++;
				break;
			case 'r' :  
				--argc;
				reorder_type = atoi((++argv)[0]);
				break;
			case 't' :  
				--argc;
				time = atof((++argv)[0]);
				break;
			case 'i' :
				--argc;
				iterations2 = atoi((++argv)[0]);
				if ((alg == W || alg == PG) && KL_PG_Flag) {
					iterations = atoi((++argv)[0]);
					--argc;
				} else
					iterations = iterations2;;
				break;
			case 'g' :  
				--argc;
				num_graphs = atoi((++argv)[0]);
				break;
			case 'k' :  
				alg = KL;
				break;
			case 's' :  
				--argc;
				seed = atol((++argv)[0]);
				break;
			case 'w' :  
				alg = W;
				break;
		}
	}
}


main(int argc, char *argv[])
{
	extern long random();	/* Unix random(3), period = 16*((2**31)-1) */
				/* every generated bit is random           */
	/*extern long lrand48();*/

	int i;
	/*long r;			/* until 12/15/93 */
	int r;

	set_defaults();
	process_args(argc, argv);
	randinit(seed);

	for (i=0; i<num_graphs; i++) {
		r = (int) MAX_INT * ( (double) random() / MAX_INT);
		/*r = random();		/* until 12/15/93		*/
		/*r = lrand48();	/* r is the seed for graph i   */
					/* until 5/93, all graphs were */
					/* generated using random().   */
					/* That function must be used  */
					/* If they are to be recreated.*/
		if ((graph_type != UNBAL_REG) && (graph_type != REG) &&
		    				 (graph_type != UNBAL_RAND))
			fprintf(outfile, "%d %.6lf %ld",num_vertices,edge_prob,r);
		else {
			if (graph_type == REG) 
			   fprintf(outfile, "%d %d %ld",num_vertices,d_reg,r);
			else if (graph_type == UNBAL_REG)
			   fprintf(outfile, "%d %d %d %ld",num_vertices,
						d_reg, d_reg2, r);
			else
			   fprintf(outfile, "%d %.6lf %.6lf %ld",num_vertices,
						edge_prob, edge_prob2, r);
		}
		if (time == 0.0) {
			if ( (alg == W) || (alg == PG) ) {
				if (KL_PG_Flag)
				    fprintf(outfile," %d 7 1", iterations2); 
				fprintf(outfile," %d 0 %d %d %d 0 ", 
					num_vertices+1, iterations, local_opt,
					kl_opt);
				if (!local_opt) fprintf(outfile," 0"); 
				fprintf(outfile," y\n"); 
			} else if (alg == KL) {
			   if (reorder_type == 7)
				fprintf(outfile," %d 7 1 0 0 %d y\n", 
						iterations, num_vertices+1);
			   else
				fprintf(outfile," %d %d 0 %d y\n", 
					iterations,reorder_type,num_vertices+1);
			}
		} else {
			if ( (alg == W) || (alg == PG) ) {
				if (KL_PG_Flag)
				    fprintf(outfile," %d 7 1", iterations2); 
				fprintf(outfile," %d 0 %.2f %d %d 0 ", 
						num_vertices+1, time,local_opt,
						kl_opt);
				if (!local_opt) fprintf(outfile," 0"); 
				fprintf(outfile," y\n"); 
			} else if (alg == KL) {
			   if (reorder_type == 7)
				fprintf(outfile," %.2f 7 1 0 0 %d y\n", 
						time, num_vertices+1);
			   else
				fprintf(outfile," %.2f %d 0 %d y\n", 
					time, reorder_type, num_vertices+1);
			}
		}
	}
	fprintf(outfile, "0\n");
	fclose(outfile);
}
