#include <stdio.h>
#include "part.h" 

extern int mode;
extern int node_cost[2][MAX_VERTICES];
extern struct adj_node *adj_list[];


int total_greedy_cost[MAX_VERTS];	/* COST if v is greedy and */
					/* its neighbors are also  */
int total_nongreedy_cost[MAX_VERTS];	/* COST if v is nongreedy  */
					/* neighs. are still greedy*/
int greedy_changes[MAX_VERTS];	/* WHEN v is greedy, how   */
					/* many neighs. changed	   */
					/* class.		   */
int nongreedy_changes[MAX_VERTS];	/* WHEN v is nongreedy, how*/
					/* many neighs. changed	   */
					/* class.		   */

int class_[MAX_VERTS];		/* STRONGLY greedy, neutral*/
					/* or strongly nongreedy   */
					/* (v unplaced)		   */
int greedy_class[MAX_VERTS];		/* THE above for each of   */
					/* v's neighs (v greedy)   */
int nongreedy_class[MAX_VERTS];	/* THE above for each of   */
					/* v's neighs,(v nongreedy)*/

int greedy_cost(int v)
{
	if (mode == MODE_MAX) {
		if (node_cost[SET_A][v] >= node_cost[SET_B][v])
			return(node_cost[SET_A][v]);
		else
			return(node_cost[SET_B][v]);
	} else {
		if (node_cost[SET_A][v] <= node_cost[SET_B][v])
			return(node_cost[SET_A][v]);
		else
			return(node_cost[SET_B][v]);
	}
}

int nongreedy_cost(int v)
{
	if (mode == MODE_MAX) {
		if (node_cost[SET_A][v] >= node_cost[SET_B][v])
			return(node_cost[SET_B][v]);
		else
			return(node_cost[SET_A][v]);
	} else {
		if (node_cost[SET_A][v] <= node_cost[SET_B][v])
			return(node_cost[SET_B][v]);
		else
			return(node_cost[SET_A][v]);
	}
}

void print_greedy_history(int v)
{
	printf(" ---------------------------------------------------\n");
	printf(" *** %-4d *** RESULTS OF GREEDY/NONGREEDY PLACEMENT\n", v);
	printf(" ---------------------------------------------------\n");
	printf(" Greedy placement: cost: %-6d, changes: %-4d\n", 
				total_greedy_cost[v], greedy_changes[v]);
	printf(" Nongreedy placement: cost: %-6d, changes: %-4d\n", 
				total_nongreedy_cost[v], nongreedy_changes[v]);
	printf(" ---------------------------------------------------\n");
}

void test_greedy_placement(int v, int *processed)
{
	int v_diff, node;
	struct adj_node *point;

	v_diff = node_cost[SET_A][v] - node_cost[SET_B][v];
	/*******************************************************************/
	/**************** pretend v has been placed greedily ***************/
	/*******************************************************************/
	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (v_diff >= 0)
				node_cost[SET_B][node]++; 
			else
				node_cost[SET_A][node]++; 
		}
	}
	
	total_greedy_cost[v] = greedy_cost(v);
	if (node_cost[SET_A][v] - node_cost[SET_B][v] >= 0)
		for (point = adj_list[v]; point; point = point->next) {
			node = point->node_no;
			if (!processed[node]) {
				greedy_class[node] = class_[node] - 1;	
				if (class_[node] /* * greedy_class[node]*/ <= 0)
					greedy_changes[v]++;
				total_greedy_cost[v] += greedy_cost(node);
			}
		}
	else
		for (point = adj_list[v]; point; point = point->next) {
			node = point->node_no;
			if (!processed[node]) {
				greedy_class[node] = class_[node] + 1;	
				if (class_[node]/* *greedy_class[node]<=*/>= 0)
					greedy_changes[v]++;
				total_greedy_cost[v] += greedy_cost(node);
			}
		}

	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (v_diff >= 0)
				node_cost[SET_B][node]--; 
			else
				node_cost[SET_A][node]--; 
		}
	}
}

void test_nongreedy_placement(int v, int *processed)
{
	int v_diff, node;
	struct adj_node *point;

	v_diff = node_cost[SET_A][v] - node_cost[SET_B][v];
	/*******************************************************************/
	/**************** pretend v has been placed nongreedily ************/
	/*******************************************************************/
	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (v_diff < 0)
				node_cost[SET_B][node]++;
			else
				node_cost[SET_A][node]++;
		}
	}
	
	total_nongreedy_cost[v] = nongreedy_cost(v);
	if (node_cost[SET_A][v] - node_cost[SET_B][v] <= 0)
		for (point = adj_list[v]; point; point = point->next) {
			node = point->node_no;
			if (!processed[node]) {
				nongreedy_class[node] = class_[node] - 1;	
				if(class_[node]/* * nongreedy_class[node]*/ <=0)
					nongreedy_changes[v]++;
				total_nongreedy_cost[v] += greedy_cost(node);
			}
		}
	else
		for (point = adj_list[v]; point; point = point->next) {
			node = point->node_no;
			if (!processed[node]) {
				nongreedy_class[node] = class_[node] + 1;	
				if(class_[node]/* *nongreedy_class[node]<=*/>=0)
					nongreedy_changes[v]++;
				total_nongreedy_cost[v] += greedy_cost(node);
			}
		}

	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (v_diff < 0)
				node_cost[SET_B][node]--;
			else
				node_cost[SET_A][node]--;
		}
	}
}

FILE *gf;

void open_greedy_outfile(int pass)
{
	extern int num, seed, sort_type;
	extern double prob;

	if (pass ==  0) {
		if ((gf = fopen("gh.out", "w")) == NULL)
			printf("Error: opening gh.out\n"), exit(9);
	} else {
		if ((gf = fopen("gh.out", "a")) == NULL)
			printf("Error: opening gh.out\n"), exit(9);
	}
	fprintf(gf,"*******************************************************\n");
	fprintf(gf,"** %d, %.6f, %d, %d %20s**\n", num, prob, seed, sort_type,"");
	fprintf(gf,"*******************************************************\n");
}


void close_greedy_outfile()
{
	fclose(gf);
}

int greedy_history(int index, int greediness, int v, int *processed)
/* index;*/		/* we are processing the <index>th vertex 	*/
/* greediness;*/	/* whether or not KL placed v greedily		*/
{
	extern int num;
	extern int vertex_loc[];
	extern FILE *outfile;

	static int ng_costs, 	    	/* # of vertices where ngcost>gcost */
		   ng_benefits,		/* # of vertices where ngbene>gbene */
	           ng_cost_matches, 	/* # of KL ng verts. w/ ngcost>gcost*/
		   ng_benefit_matches;	/* # of KL ng verts. w/ ngbene>gbene*/
	int i, node, unproc_neighs=0;
	struct adj_node *point;

	for (i=0; i<num; i++) {
		class_ [i] = greedy_class[i] = nongreedy_class[i] = 0;
		greedy_changes[i] = nongreedy_changes[i] = 0;
		total_greedy_cost[i] = total_nongreedy_cost[i] = 0;
	}
	/*** undo the changes made in postproc by placing v ***/
	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (vertex_loc[v])
				node_cost[SET_A][node]--;
			else
				node_cost[SET_B][node]--;
		}
	}


	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
		   class_[node]=node_cost[SET_A][node]-node_cost[SET_B][node];
		   unproc_neighs++;
		}
	}
	test_greedy_placement(v, processed);
	test_nongreedy_placement(v, processed);

	/*** remake the changes made in postproc by placing v ***/
	for (point = adj_list[v]; point; point = point->next) {
		node = point->node_no;
		if (!processed[node]) {
			if (vertex_loc[v])
				node_cost[SET_A][node]++;
			else
				node_cost[SET_B][node]++;
		}
	}

	if (greediness == NONGREEDY) 
		fprintf(gf, "**");
	if (greedy_changes[v] < nongreedy_changes[v]) {
		ng_benefits++;
		fprintf(gf,"%-4d*: u_neighs: %d, g_chngs: %d, ng_chngs: %d",
			v,unproc_neighs,greedy_changes[v],nongreedy_changes[v]);
		fprintf(gf,", g_cost: %d,ng_cost:, %d",
			total_greedy_cost[v], total_nongreedy_cost[v]);
	} else {
		fprintf(gf,"%-4d: u_neighs: %d, g_chngs: %d, ng_chngs: %d",
			v,unproc_neighs,greedy_changes[v],nongreedy_changes[v]);
		fprintf(gf,", g_cost: %d,ng_cost:, %d",
			total_greedy_cost[v], total_nongreedy_cost[v]);
	}
	if (total_greedy_cost[v] < total_nongreedy_cost[v]) {
		fprintf(gf,"***\n");
		ng_costs++;
	} else
		fprintf(gf,"\n");

	if ((greediness==NONGREEDY) && (greedy_changes[v]<nongreedy_changes[v]))
		ng_benefit_matches++;
	if ((greediness==NONGREEDY) && 
				(total_greedy_cost[v]<total_nongreedy_cost[v]))
		ng_cost_matches++;
	if (index == num-1) {
		fprintf(outfile,"------------------------------------------\n");
		fprintf(outfile,"%%nongreedy benefiting verts that KL placed ");
		fprintf(outfile, "nongreedily: %%%.1f\n", 
			((double) ng_benefit_matches/ (double) ng_benefits)*100); 
		fprintf(outfile, "%%nongreedy cost verts that KL placed "); 
		fprintf(outfile, "nongreedily: %%%.1f\n", 
			((double) ng_cost_matches/ (double) ng_costs)*100); 
	}

	if ((mode==MODE_MAX && total_greedy_cost[v]>=total_nongreedy_cost[v])||
	    (mode==MODE_MIN && total_greedy_cost[v]<=total_nongreedy_cost[v]))
		return(TRUE);
	else
		return(FALSE);
}
