#include "part.h"
#include "backtrack.h"

typedef struct {
	int size;
	int hp[MAX_VERTICES+1];
	int locs[MAX_VERTICES];
} HEAP, *HEAPPTR;


HEAP	  heap;

#define PARENT(x) 	( (int) (x / 2) )
#define LEFT(x)   	( x * 2 )
#define RIGHT(x)  	( x * 2 + 1 )
#define DIFF_DEG_COMPARE(x, y)   ( \
		   ( (abs_diff[heap.hp[x]] < abs_diff[heap.hp[y]]) || \
		     ((abs_diff[heap.hp[x]] == abs_diff[heap.hp[y]]) && \
		      (deg[heap.hp[x]] < deg[heap.hp[y]])) ) )

int abs_diff[MAX_VERTICES];
int deg[MAX_VERTICES];

int heap_maximum()
{
	return(heap.hp[1]);
}

heap_insert(new)
{
	int 	parent, child, temp;

	heap.size++;
	heap.hp[heap.size] = new;
	heap.locs[new] = heap.size;
	child = heap.size;
	parent = PARENT(child);
	if (parent == child) return;
	while (parent >= 1) {
		if (DIFF_DEG_COMPARE(parent, child)) {
			temp = heap.hp[parent];
			heap.hp[parent] = heap.hp[child];
			heap.hp[child] = temp;
			heap.locs[temp] = child;
			heap.locs[heap.hp[parent]] = parent;
			
			child = parent;
			parent = PARENT(parent);
		} else
			parent = 0;
	}
}

int heap_extract_max(target)
int target;
{
	int i, max, parent, child, temp, tempptr = 0;
	static int temparray[MAX_VERTICES];

	do {
		if (heap.size == 0)
			return(BOS);
	
		max = heap.hp[1];

		heap.hp[1] = heap.hp[heap.size];
		heap.locs[heap.hp[heap.size]] = 1;
		heap.size--;
		parent = 1;
		child = 2;
		while (child <= heap.size) {
			if ((child < heap.size) && DIFF_DEG_COMPARE(child, child+1)) 
				child++;
			if (DIFF_DEG_COMPARE(parent, child)) {
				temp = heap.hp[parent];
				heap.hp[parent] = heap.hp[child];
				heap.hp[child] = temp;
				heap.locs[temp] = child;
				heap.locs[heap.hp[parent]] = parent;
			
				parent = child;
				child = LEFT(child);
			} else
				break;
		}
		heap.locs[max] = 0;
		temparray[tempptr++] = max;
	} while ((target != WILDCARD) && (target != max));
	if (target != WILDCARD)
		for (i=0; i<tempptr-1; i++)
			heap_insert(temparray[i]);
	return(max);
}

heap_change_priority(node)
int 	node;
{
	int 	loc = 	 heap.locs[node],
		parent = PARENT(loc),
		child =  LEFT(loc),
		temp;
	
	if ( (parent > 0) &&
	     (DIFF_DEG_COMPARE(parent, loc)) ) {

		/** increase priority **/
		while (parent >= 1) {
	     	  if (DIFF_DEG_COMPARE(parent, loc)) {
			temp = heap.hp[parent];
			heap.hp[parent] = heap.hp[loc];
			heap.hp[loc] = temp;
			heap.locs[temp] = loc;
			heap.locs[heap.hp[parent]] = parent;
			
			loc = parent;
			parent = PARENT(parent);
		  } else
			break;      
		}
	} else if ( (child <= heap.size) &&
		    ( (DIFF_DEG_COMPARE(loc, child)) ||
		      (DIFF_DEG_COMPARE(loc, child+1))) ) {

		/** decrease priority **/
		while (child <= heap.size) {
		  if ( (child < heap.size) && DIFF_DEG_COMPARE(child, child+1)) 
			child++;
		  if (DIFF_DEG_COMPARE(loc, child)) {
			temp = heap.hp[loc];
			heap.hp[loc] = heap.hp[child];
			heap.hp[child] = temp;
			heap.locs[temp] = child;
			heap.locs[heap.hp[loc]] = loc;
			
			loc = child;
			child = LEFT(child);
		  } else
			break;                 
		}
	}
}

print_heap()
{
	int i;

	printf("-----------------------------------------\n");
	printf("-- Heap Contents ------------------------\n");
	printf("-----------------------------------------\n");
	for (i=1; i <= heap.size; i++)
		printf("%d ",heap.hp[i]);
	printf("\n");
	for (i=1; i <= heap.size; i++)
		printf("%d ",abs_diff[heap.hp[i]]);
	printf("\n");
	for (i=1; i <= heap.size; i++)
		printf("%d ",deg[heap.hp[i]]);
	printf("\n");
	printf("-----------------------------------------\n");
}

/*
main()
{
	int num = 0, val, degree, i;
	char c[1];

	while (num >=0) {
		scanf("%s %d %d %d", c, &num, &val, &degree);
		abs_diff[num] = val;
		deg[num] = degree;
		if (c[0] == 'i') {
			heap_insert(num);
			print_heap();
		} else if (c[0] == 'e') {
			i = heap_extract_max(num);
			printf("EXTRACT_MAX: %d\n", i);
			print_heap();
		} else if (c[0] == 'c') {
			abs_diff[num] = val;
			heap_change_priority(num);
			print_heap();
		} else
			printf("scanf barfed\n");
	}
}
*/
