#include <stdio.h>
#include "part.h"

#define CUTS 0
#define NONCUTS 1
#define BUCKET_NUM(cost)  ((cost) + Num_Pos_Buckets)

typedef struct tlist_holder {
	int next_slot;		/** next available vertex index            **/
	double best_value;	/** of this list, the best cut or quotient **/
				/** value.				   **/
	int length;		/** how many elements 			   **/
	int backup_num;		/** how many elements must be deleted off  **/
				/** the front of the list to get to that   **/
				/** best value				   **/
				/** this allows for negative flip cost     **/
				/** increments although some versions of   **/
				/** PO don't do this.                      **/
	int sum_costs;		/** of all vertices in list                **/
	int sum_cuts;		/** number of cut edges induced            **/
	int sum_noncuts;	/** number of noncut edges induced         **/
	TLIST l[MAX_VERTICES];	/** the actual list			   **/
} GLIST, *GLISTPTR;

void best_start(int, int, int *bnum, LISTPTR*, LISTPTR*);
void insert_glist(GLISTPTR glist, LISTPTR, int, int);
void move_to_front(int node, int neighbor);
void init_glist(GLISTPTR glist);
void save_and_postproc(int, int, int, int *, int *);
int compute_costs();
void save_state_if_improvement(int local);
void kl_pass(int);
void print_iteration(int, int, int, int, int);


extern int Dont_Compute_Costs;
extern LISTPTR *buckets[2];

extern int Num_Pos_Buckets;
extern int num;
extern int currentcost;
extern int num_verts[];				/* from bk     */
extern int num_vertices[];			/* from bucket */
extern int vert_loc[];
extern int node_cost[2][MAX_VERTICES];
extern int hold_num_verts[], hold_loc[], hold_ncost[];
extern struct adj_node *adj_list[MAX_VERTICES];
extern FILE *outfile;

int hours;
double last_time;

double Ave_Path_Length;
int Num_Augmenting_Paths;
int Total_Flip_Cost;
int Total_Path_Length;
int bucket_search_incr;
int bucket_search_starting_limit;
int bucket_search_limit;

GLIST gl;
GLISTPTR glist = &gl;
/***************************************************************************/
/** NOTE:  this file will be appalling to anybody familiar with flexible  **/
/**        program design.  some routines which should take GLISTPTRs as  **/
/**	   parameters instead use the global "glist."  This is because we **/
/**	   only deal with exactly one GLIST.  Frequently called routines  **/
/**	   are more efficient without the parameter overhead.  The whole  **/
/** 	   GLIST type could be very nicely written in C++, but would run  **/
/**	   slower (probably).						  **/
/***************************************************************************/

int glist_backup()
{
	extern GLISTPTR glist;
	extern LISTPTR nodes;
	extern int mode;
	int max = (mode == MODE_MAX);

	if (Dont_Compute_Costs) {
		if ((max && (nodes[glist->l[glist->next_slot-1].v].cost<-1)) ||
		    (!max&& (nodes[glist->l[glist->next_slot-1].v].cost>1)))
			return TRUE;
	} else {
		if (glist->backup_num > 0)
			return TRUE;
	}
	return FALSE;
}

TLISTPTR glist_remove_first(GLISTPTR glist)
{
	extern LISTPTR nodes;
	TLISTPTR temp;

	if (--glist->next_slot < 0)
		return NULL;
	
	glist->length--;
	temp = &glist->l[glist->next_slot];
	if (!Dont_Compute_Costs) {
		glist->backup_num--;
		glist->sum_costs -= nodes[temp->v].cost;
		glist->sum_cuts -= temp->cuts;
		glist->sum_noncuts -= temp->noncuts;
	}
	return temp;
}

void bucket_move_vertex(LISTPTR node)
{
	extern LISTPTR nodes;
	extern int currentminsize;

	int loc;
	struct adj_node *point;
	LISTPTR temp_node;

	num_vertices[node->loc]--;
	num_vertices[!node->loc]++;

	currentcost += node->cost;
	currentminsize = MIN(num_vertices[SET_A], num_vertices[SET_B]);
	loc = !node->loc;
	for (point = adj_list[node->node_no]; point; point = point->next) {
		if (nodes[point->node_no].lock == NOT_LOCKED) {
			temp_node = &nodes[point->node_no];
			if (loc != temp_node->loc) 
				move_to_bucket(temp_node->loc, 
				               temp_node->cost-2,temp_node);
			else 
		   		move_to_bucket(temp_node->loc,
				   	       temp_node->cost+2,temp_node);
		} /*else
			printf("AHA******************%d\n", point->node_no);*/
	}
	node->loc = loc;
}

void init_local_opt(int *locs, int *num_verts)
{
	extern LISTPTR nodes;
	extern int currentcost, currentminsize;

	int i;

	num_vertices[0] = num_verts[0];
	num_vertices[1] = num_verts[1];
	for (i=0; i<num; i++)
		nodes[i].loc = locs[i];
	currentcost = compute_costs();
	currentminsize = MIN(num_vertices[0], num_vertices[1]);
	reset_buckets(node_cost);
}


void local_opt()
{
	extern int mode;
	extern int output_short;
	extern int Paths;

	extern LISTPTR bucket_extract_max();
	extern LISTPTR bucket_extract_min();
	int count = 0;
	LISTPTR node;

	if (mode == MODE_MAX) 

	for (node=bucket_extract_max(NOT_LOCKED);node->cost>0;
					 node=bucket_extract_max(NOT_LOCKED)){
		bucket_move_vertex(node);
		insert_to_bucket(node->loc, -node->cost, node);
		count++;
	}
	
	else /* MINIMIZE */

	for (node=bucket_extract_min(NOT_LOCKED);node->cost<0;
					 node=bucket_extract_min(NOT_LOCKED)){
		bucket_move_vertex(node);
		insert_to_bucket(node->loc, -node->cost, node);
		count++;
	}
	
	insert_to_bucket(node->loc, node->cost, node);
	if (!output_short && Paths != BACKTRACK) {
	    fprintf(outfile,"AFTER LOCAL OPTIMIZATION (%d moves):\n", count);
	    fprintf(outfile,"%-6d%-11s%-20s%-11s%-11s%-6d%-4d\n",
		currentcost, "", "", "", "",
               	num_vertices[SET_A],num_vertices[SET_B]);
	}
	save_state_if_improvement(AUGMENT);
}

void balance_partitioning()
{
	extern int mode, currentminsize;
	extern int output_short;

	extern LISTPTR extract_min_from_side();
	int count = 0, side, dummy;
	int next_minsize;
	double cur_quot, next_quot;
	LISTPTR node;

	side = (num_vertices[0] <= num_vertices[1]) ? 1 : 0;

	node=extract_min_from_side(side, NOT_LOCKED);
	while (node) {
		cur_quot = currentcost / (double) currentminsize;
		next_minsize = MIN(num_vertices[side]-1, num_vertices[!side]+1);
		next_quot = (currentcost+node->cost) / (double) next_minsize;
		if (cur_quot <= next_quot)
			break;
		bucket_move_vertex(node);
		insert_to_bucket(node->loc, -node->cost, node);
		count++;
		node=extract_min_from_side(side, NOT_LOCKED);
	}
	insert_to_bucket(node->loc, node->cost, node);
	if (!output_short) {
	    fprintf(outfile,"AFTER BALANCING (%d moves):\n", count);
                   print_iteration(0,0,currentcost,
                                                   currentminsize, 0);
	}
	save_and_postproc(mode, 0, 0, &dummy, &dummy);
}

int kl_maximize() 
{
	extern int currentcost;
	int iter=0, oldcost = VERY_LOW;

	while (oldcost < currentcost) {
		oldcost = currentcost;
		iter++;
		kl_pass(0);
	}
	return(iter);
}

int kl_min_quot() 
{
	extern int currentcost, currentminsize;
	int iter=0, oldcost = VERY_HIGH, oldminsize = 1;

        while (1) {
              iter++;
              kl_pass(0);
              if ( ((double) oldcost / (double) oldminsize) <=
                             ((double) currentcost / (double) currentminsize))
                     break;
              oldcost = currentcost;
              oldminsize = currentminsize;
        }
	return iter;
}

int kl_minimize() 
{
	int iter=0, oldcost = VERY_HIGH;

	while (oldcost > currentcost) {
		oldcost = currentcost;
		iter++;
		kl_pass(0);
	}
	return(iter);
}


void kl_augment(int *locs)
{
	extern int output_short;
	extern int mode;
	extern int Paths;
	extern LISTPTR nodes;
	int i, iter;

	for (i=0; i<num; i++)
		nodes[i].loc = locs[i];
	if (mode == MODE_MAX)
		iter = kl_maximize();
	else if (mode == MIN_QUOTIENT)
		iter = kl_min_quot();
	else
		iter = kl_minimize();
	if (!output_short && Paths != BACKTRACK) {
	    fprintf(outfile,"AFTER KL AUGMENTATION (%d iter):\n", iter);
	    fprintf(outfile,"%-6d%-11s%-20s%-11s%-11s%-6d%-4d\n",
		currentcost, "", "", "", "",
               	num_vertices[SET_A],num_vertices[SET_B]);
	}
	save_state_if_improvement(AUGMENT);
}

int path(int v)
{	
	extern GLISTPTR glist;
	extern LISTPTR nodes;
	extern int Strict_Trees;
	extern int mode;
	int edges[2] = { 0, 0 };
	int i, i0, i1, side, new_side;
	int max;
	int mode_mult;
	/** int total_cost, sum_cuts, sum_noncuts;  for allowing new f.c.i. **/

	if (glist->next_slot <= 0)
		return FALSE;
	max = (mode == MODE_MAX);
	mode_mult = (mode == MODE_MAX) ? 1 : -1;
	side =  nodes[glist->l[glist->next_slot-1].v].loc;
	new_side =  nodes[v].loc;
	i0 = (side == new_side);
	i1 = !i0;

	for (i=glist->next_slot-1; i>=0; i--) {
		edges[i0] += adj(v, glist->l[i].v);
		if (--i >= 0) 
			edges[i1] += adj(v, glist->l[i].v);
		if (Strict_Trees) 
			if (edges[max] > 0) /* || edges[!max]>1)*/
				return FALSE;
	}
	/*total_cost = nodes[v].cost + glist->sum_costs;
	sum_cuts =  edges[CUTS] + glist->sum_cuts;
	sum_noncuts =  edges[NONCUTS] + glist->sum_noncuts;*/
	if  (((max   && (1 <= edges[!max])) ||
	      (!max  && (MIN(1,glist->length-1) <= edges[!max]))) &&
	     ((mode_mult *(nodes[v].cost+2*(edges[CUTS]-edges[NONCUTS])))
									>=0)){
	     /*((mode_mult *(total_cost+2*(sum_cuts-sum_noncuts)))
									>=0)){*/
		insert_glist(glist, &nodes[v], edges[CUTS], edges[NONCUTS]);
		return TRUE;
	}
	return FALSE;
}

int path2(int v)
{	
	extern GLISTPTR glist;
	extern LISTPTR nodes;
	extern int Strict_Trees;
	extern int mode;
	int i;
	int current_index = glist->next_slot-1;
	int max_flag = (mode == MODE_MAX);
	TLISTPTR tlist=&glist->l[glist->next_slot-1], t1;

	if (glist->next_slot <= 0)
		return FALSE;
	
	if (max_flag) {
		if (!adj(v, tlist->v))
			return FALSE;
	} else {
		if ((t1 = &glist->l[--current_index])) {
			if (!adj(v, t1->v))
				return FALSE;
		}
		current_index = glist->next_slot-1;
		for (i=current_index; i>=0; i-=2) {
			t1 = &glist->l[current_index];
			if (adj(v, t1->v))
				return FALSE;
		}
	}
	if (!Strict_Trees) {
		insert_glist(glist, &nodes[v], 0, 0);
		return TRUE;
	}
	current_index = glist->next_slot - (1 + max_flag);
	for (i=current_index; i>=0; i-=2) {
		t1 = &glist->l[current_index];
		if (adj(v, t1->v))
			return FALSE;
	}
	insert_glist(glist, &nodes[v], 0, 0);
	return TRUE;
}


LISTPTR find_next_match(int side)
{
	/* extern GLISTPTR glist; */

	int search = TRUE, bnum[2];
	LISTPTR p, p2;
	int  (*path_f)(int);

	if (Dont_Compute_Costs)
		path_f = path2;
	else
		path_f = path;

	/*limit = BUCKET_NUM(-1 * (2 * glist->length/2 + 2)); */
	/*limit = BUCKET_NUM(2 * glist->length/2 + 2);*/

	best_start(side, bucket_search_limit, bnum, &p, &p2);

	while (search) {
		if (!p) {
			if (bnum[side] != bucket_search_limit) {
				bnum[side] = bnum[side] + bucket_search_incr;
				p = buckets[side][bnum[side]]->next;
			} else
				search = FALSE;
		}
		for (; p; p=p->next) {
			if ((p->lock == NOT_LOCKED) && path_f(p->node_no)){
				p->lock = LOCKED;
				return p;
			}
		}
	}
	return NULL;
}

LISTPTR find_next_match2(int side)
{
	extern GLISTPTR glist;
	extern LISTPTR nodes;
	extern int mode;
	TLISTPTR tlist;
	LISTPTR p;
	struct adj_node *point;

	if (mode != MODE_MAX) {
		if (glist->next_slot > 1)
			tlist = &glist->l[glist->next_slot-2];
		else
			return NULL;
	} else {
		tlist = &glist->l[glist->next_slot-1];
	}
	for (point=adj_list[tlist->v]; point; point=point->next) {
		p = &nodes[point->node_no];	
		if ((p->loc == side) && (p->lock == NOT_LOCKED) && 
		    path(p->node_no)){
			p->lock = LOCKED;
			return p;
		}
	}
	return NULL;
}

GLISTPTR augmenting_path(LISTPTR p, LISTPTR p2)
{
	extern GLISTPTR glist;
	extern int mode;
	extern int Long_Augmenting_Paths;
	extern LISTPTR nodes;
	TLISTPTR ttemp;
	LISTPTR temp;
	int side;
	LISTPTR (*find_match)(int);

	if (Long_Augmenting_Paths)
		find_match = find_next_match2;
	else
		find_match = find_next_match;

	init_glist(glist);
	insert_glist(glist, p, 0, 0);
	p->lock = LOCKED;
	if ((mode != MODE_MAX) && p2) {
		insert_glist(glist, p2, adj(p->node_no,p2->node_no), 0);
		p2->lock = LOCKED;
		side = !p2->loc;
	} else
		side = !p->loc;
	while ((temp = find_match(side))) 
		side = !temp->loc;
	while (glist_backup()) {
		ttemp = glist_remove_first(glist);
		nodes[ttemp->v].lock = NOT_LOCKED;
	}
	return glist;
}

void print_path(GLISTPTR glist)
{	
	extern LISTPTR nodes;
	TLISTPTR temp;

	int i, flip_cost;

	if (glist->next_slot <= 0) return;

	flip_cost=glist->sum_costs+ 2*(glist->sum_cuts - glist->sum_noncuts);
	printf("AUGMENTING PATH (gain:%d): ", flip_cost);
	for (i=glist->next_slot-1; i>=0; i--) {
		temp = &glist->l[i];
		printf("%d(%d,%d,%d) ",temp->v, nodes[temp->v].cost,
			temp->cuts, temp->noncuts);
		nodes[temp->v].lock = NOT_LOCKED;
	}
	printf("\n");
}

void dispose_path(GLISTPTR glist)
{
	extern LISTPTR nodes;
	TLISTPTR temp;

	while ( (temp=glist_remove_first(glist)) != NULL) { 
		nodes[temp->v].lock = NOT_LOCKED;
	}
}
 
void flip_path(GLISTPTR glist)
{
	extern int output_short, Paths, mode, alg_type;
	extern LISTPTR nodes;
	TLISTPTR temp;
	int flip_cost;
	int i, v, dummy=currentcost, *c=&dummy;

	if (glist->next_slot <= 0) return;

	flip_cost=glist->sum_costs+ 2*(glist->sum_cuts - glist->sum_noncuts);
	Total_Flip_Cost += flip_cost;

	for (i=0; i<glist->next_slot; i++) 
		nodes[glist->l[i].v].lock = NOT_LOCKED;

	while ( (temp=glist_remove_first(glist)) != NULL) { 
		v = temp->v;
		delete_from_bucket(&nodes[v]);
		bucket_move_vertex(&nodes[v]);
		insert_to_bucket(nodes[v].loc, -nodes[v].cost, &nodes[v]);
	}
	if (!output_short && Paths != BACKTRACK) {
	    printf("AFTER AUGMENTING PATHS OPTIMIZATION :\n");
	    printf("%-6d%-11s%-20s%-11s%-11s%-6d%-4d\n",
		currentcost, "", "", "", "",
               	num_vertices[SET_A],num_vertices[SET_B]);
	}
	if (alg_type == BACKTRACKING)
		save_state_if_improvement(AUGMENT);
	else if (alg_type == ALG_TYPE_KL) 
		save_and_postproc(mode, 0, 0, c, c);
}

void init_glist(GLISTPTR glist)
{
	extern int mode;

	if (mode == MODE_MAX)
		glist->best_value = 0.0;
	else
		glist->best_value = (double) MAX_INT;
	glist->next_slot = 0;
	glist->length = 0;
	glist->backup_num = 0;
	glist->sum_costs = 0;
	glist->sum_cuts = 0;
	glist->sum_noncuts = 0;
}

void insert_glist(GLISTPTR glist, LISTPTR p,int cuts, int noncuts)
{
	extern LISTPTR nodes;
	extern int mode;
	extern int currentcost, currentminsize;
	double new_cost, new_quot;
	int side, new_minsize; 

	glist->length++;
	if (Dont_Compute_Costs) {
		glist->l[glist->next_slot++].v = p->node_no; 
	} else {
		TLISTPTR t;

		glist->sum_costs = p->cost + glist->sum_costs;
		glist->sum_cuts = cuts + glist->sum_cuts;
		glist->sum_noncuts = noncuts + glist->sum_noncuts;
		new_cost = currentcost + 
		   	(glist->sum_costs + 
				(2 * (glist->sum_cuts - glist->sum_noncuts)));
		if (mode == MODE_MAX) {
			if (new_cost >= glist->best_value) {
				glist->best_value = new_cost;
				glist->backup_num = 0;
			} else
				glist->backup_num++;
		} else {
			if ((glist->length % 2) == 0)
		   		new_minsize = currentminsize;
			else {
		   		side = nodes[p->node_no].loc;
		   		new_minsize =MIN(num_vertices[side]-1,
						 num_vertices[!side]+1);
			}
			if ( (new_quot=new_cost/new_minsize) <= 
			      glist->best_value) {
				glist->best_value = new_quot;
				glist->backup_num = 0;
			} else
				glist->backup_num++;
		}
		t = &glist->l[glist->next_slot++];
		t->v = p->node_no; 
		t->cuts = cuts; 
		t->noncuts = noncuts;
	}  
}

int improvement(GLISTPTR glist)
{
	extern LISTPTR nodes;
	extern int currentcost, currentminsize;
	extern int mode;
	int max_flag = (mode==MODE_MAX);

	if (Dont_Compute_Costs) {
	  if (glist->length > 1) return TRUE;
	  if (glist->length == 1) {
		LISTPTR t = &nodes[glist->l[glist->next_slot-1].v];
		if (max_flag) {
			if (t->cost >= 0)
				return TRUE;
		} else {
			int new_minsize, side;
			if ((glist->length % 2) == 0)
		   		new_minsize = currentminsize;
			else {
		   		side = t->loc;
		   		new_minsize =MIN(num_vertices[side]-1,
						 num_vertices[!side]+1);
			}
			if ( ((currentcost+t->cost) / (double)new_minsize) <= 
			      (currentcost / (double)currentminsize)) 
				return TRUE;
	  	}
	  }
	} else {
	  if ( (max_flag && (glist->best_value > currentcost)) ||
	       (!max_flag&& 
	        (glist->best_value < ((double) currentcost/currentminsize)))) 
			return TRUE;
	}
	return FALSE;
}

void preprocess_adjlist(int side)
/********************************* currently unused ********/
{
	int search[2] = { TRUE, TRUE }, 
	    bnum[2], backup_bnum;
	LISTPTR p, p1;

	best_start(side, bucket_search_limit, bnum, &p, &p1);
	backup_bnum = bnum[!side];
	while (search[side]) {
		if (!p) {
			if (bnum[side] != bucket_search_limit) {
				bnum[side] = bnum[side] + bucket_search_incr;
				p = buckets[side][bnum[side]]->next;
			} else 
				search[side] = FALSE;
		}
		for (; p; p=p->next) {
			bnum[!side] = backup_bnum;
			p1 = buckets[!side][bnum[!side]]->next;
			while (search[!side]) {
				if (!p1) {
					if (bnum[!side] == bucket_search_limit){
					   bnum[!side] = bnum[!side] + 
							bucket_search_incr;
				 	   p1=buckets[!side][bnum[!side]]->next;
					} else
						search[!side] = FALSE;
				}
				for (; p1; p1=p1->next) 
					move_to_front(p->node_no,p1->node_no);
			}
		}
	}
}

void best_start(int side, int search_limit, int bnum[2], 
		LISTPTR *best_of_side, LISTPTR *best_of_other_side)
{
	extern int mode;
	extern int Current_Max_Bucket[2];
	extern int Current_Min_Bucket[2];
	extern int Long_Augmenting_Paths;

	int search = TRUE;

	*best_of_side = *best_of_other_side = NULL;

	if (mode == MODE_MAX) {
		bnum[side]=(Current_Max_Bucket[side]>BUCKET_NUM(0)) 
				? Current_Max_Bucket[side] : BUCKET_NUM(0);
		bnum[!side]=(Current_Max_Bucket[!side]>BUCKET_NUM(0)) 
				? Current_Max_Bucket[!side] : BUCKET_NUM(0);
	} else {
		bnum[side] =(Current_Min_Bucket[side]<BUCKET_NUM(0)) 
				? Current_Min_Bucket[side] : BUCKET_NUM(0);
		bnum[!side] =(Current_Min_Bucket[!side]<BUCKET_NUM(0)) 
				? Current_Min_Bucket[!side] : BUCKET_NUM(0);
	}
	*best_of_side = buckets[side][bnum[side]]->next;

	while (search) {
		if (!*best_of_side) {
			bnum[side] = bnum[side] + bucket_search_incr;
			*best_of_side = buckets[side][bnum[side]]->next;
			if (bnum[side] == search_limit)
				search = FALSE;
		} else break;
	}
	if ((mode == MODE_MAX) || (!Long_Augmenting_Paths))
		return;
	*best_of_other_side = buckets[!side][bnum[!side]]->next;
	search = TRUE;
	while (search) {
		if (!*best_of_other_side) { 
			bnum[!side] = bnum[!side] + bucket_search_incr;
			*best_of_other_side = buckets[!side][bnum[!side]]->next;
			if (bnum[!side] == search_limit)
				search = FALSE;
		} else break;
	}
}

void path_from_side(int side)
{
	extern GLISTPTR glist;

	int 	search = TRUE; 
	int     bnum[2];
	LISTPTR p, p2;

	best_start(side, bucket_search_starting_limit, bnum, &p, &p2);

	while (search) {
		if (!p) {
			if (bnum[side] != bucket_search_starting_limit) {
				bnum[side] = bnum[side] + bucket_search_incr;
				p = buckets[side][bnum[side]]->next;
			} else
				search = FALSE;
		}
		for (; p; p=p->next) {
			glist = augmenting_path(p, p2);
			if (improvement(glist)) {
				print_path(glist);
				Total_Path_Length += glist->next_slot;
				Num_Augmenting_Paths++;
				flip_path(glist);
				return;			/* break */
			} else
				dispose_path(glist);
		}
	}
}

void find_augmenting_paths()
{
	extern int mode;
	extern int Output_Sng_Distfile;
	extern int Heuristic_Only;
	extern LISTPTR nodes;
	extern int currentcost, currentminsize;
        extern int tbestcost, tbestminsize;
        extern int bk_cost, bk_minsize;
	extern double tottime;
	extern double cpu_seconds;
	/* extern FILE *tfile; */
	extern double time_;

	int i, side=0;
	int lastcost1, lastcost2;
	
	for (i=0; i<num; i++)
		nodes[i].lock = NOT_LOCKED;
	bucket_search_incr = (mode==MODE_MAX) ? -1 : 1; 
	bucket_search_starting_limit = (mode == MODE_MAX) 
					? BUCKET_NUM(-1) : BUCKET_NUM(1);
					/*? BUCKET_NUM(Num_Pos_Buckets * -1) 
					: BUCKET_NUM(Num_Pos_Buckets);*/
	bucket_search_limit = (mode == MODE_MAX) ? BUCKET_NUM(-2) : 
				        BUCKET_NUM(2);
	/**starting_limit = BUCKET_NUM(-1 * Num_Pos_Buckets); **/
	/**starting_limit = BUCKET_NUM(Num_Pos_Buckets);      **/

	if (mode == MODE_MAX) {
		lastcost1 = lastcost2 = 0;
		while ((lastcost1 < currentcost) || (lastcost2 < currentcost)) {
			lastcost2=lastcost1;
			lastcost1=currentcost;
			path_from_side(side);
			side = !side;
			if (Heuristic_Only) {
				time_ = log_time();
				tottime += time_;
				if (tottime >= cpu_seconds)
					break;
				if (tottime >= (last_time + 3600.0)) {
					hours++;
					last_time = tottime;
					/*fprintf(tfile,"hours: %d, quot: %d\n",
						hours, 
						bk_cost);
					fflush(tfile);*/
				}
			}
		}
	} else {
		double 	lastcost1, lastcost2, lastquot1, lastquot2; 
		double currentquot = (double) currentcost/ currentminsize;
		lastcost1=lastcost2=lastquot1=lastquot2= (double) MAX_INT;
		while ( ((lastcost1>currentcost) || (lastcost2>currentcost)) ||
		        ((lastquot1>currentquot) || (lastquot2>currentquot))) {
		    lastquot2=lastquot1;
		    lastquot1=currentquot;
		    lastcost2=lastcost1;
		    lastcost1=currentcost;
		    path_from_side(side);
		    side = !side;
		    currentquot = (double) currentcost/ currentminsize;
		    if (Heuristic_Only) {
		    	time_ = log_time();
		    	tottime += time_;
		    	if (tottime >= cpu_seconds)
				break;
			if (tottime >= (last_time + 3600.0)) {
				hours++;
				last_time = tottime;
				/*fprintf(tfile,"hours: %d, quot: %f\n",
					hours, 
					(double)bk_cost/bk_minsize);*/
			}
		    }
		}
	}
	if (Output_Sng_Distfile) {
		tbestcost = bk_cost;
		tbestminsize = bk_minsize;
		post_processing(TRUE, 0, 0);
	}
	Ave_Path_Length = (Num_Augmenting_Paths  > 0) ? 
			Total_Path_Length/(double) Num_Augmenting_Paths : 0.0;
}
