#include<stdio.h>
#include "part.h"

#define gp(n) (n + 1)				/* graphpack verts. st.w/1 */

int adj(int, int);

extern int num, numedges;
extern int strongly_ng;				/* # nongreedy vertices    */
extern int vertices[]; 
extern struct adj_node *adj_list[MAX_VERTICES];
extern VLIST nongreedy_vertices[];

static int total_cost;

FILE *gpfile, *gpinfo;

void init_gp_info()
{
	if ((gpinfo = fopen("gpinfo", "w")) == NULL)
		printf("Error: can't open graphpack info file\n"),exit(56);
	fprintf(gpinfo,"Vertex order  Cuts\n");
	total_cost = 0;
}

void gp_info(int v, int s_deg, int side, int costa, int costb)
{

	fprintf(gpinfo, "%12d, %4d, %4d", gp(v), costa, costb);
	if (!side)
		total_cost += costa;
	else
		total_cost += costb;
	fprintf(gpinfo, "%6d   %2d\n", total_cost, side);
	fflush(gpinfo);
}

void output_gp()
/***********************************************************************/
/** output_gp()		Finish up the gpinfo file by printing nongr.  **/
/**			verts., then create the graphpack data file.  **/
/***********************************************************************/
{
	int i, start;
	struct adj_node *temp;

	if ((gpfile = fopen("../gpfile", "w")) == NULL)
		printf("Error: can't open graphpack output file\n"),exit(54);

	fprintf(gpinfo," Nongreedy Vertices: ");
	for (i=0; i<strongly_ng; i++)
		fprintf(gpinfo, ",%d", gp(nongreedy_vertices[i].v));
	fprintf(gpinfo, "\n");

	fprintf(gpfile, "VERTICES:{1");
	for (i=1; i<num; i++)
		fprintf(gpfile, ",%d", gp(i));
	fprintf(gpfile, "}\n");

	fprintf(gpfile, "VWEIGHTS:*1");
	for (i=1; i<num; i++)
		fprintf(gpfile, ",%d", gp(i));
	fprintf(gpfile, "*\n");

	fprintf(gpfile, "EDGES:{");
	start = FALSE;
	for (i=0; i<num; i++) {
		if (start) {
			for (temp=adj_list[i]; temp; temp= temp->next)
				fprintf(gpfile,", {%d,%d}",gp(i),
						  gp(temp->node_no));
		} else if (adj_list[i]) {
		    fprintf(gpfile,"{%d,%d}",gp(i),gp(adj_list[i]->node_no));
		    for (temp=adj_list[i]->next; temp; temp= temp->next)
				fprintf(gpfile,", {%d,%d}",gp(i),
						  gp(temp->node_no));
		    start= TRUE;
		} 
	}
	fprintf(gpfile,"}\n");

	fprintf(gpfile,"EWEIGHTS:*");
	for(i=0; i<numedges-1; i++)
		fprintf(gpfile, ",");
	fprintf(gpfile,"*\n");

	fclose(gpfile);
	fclose(gpinfo);
}

void output_ng()
/***********************************************************************/
/** output_gp()		Finish up the gpinfo file by printing nongr.  **/
/**			verts., then create the graphpack data file.  **/
/***********************************************************************/
{
	int i, j, start, edges = 0;

	if ((gpfile = fopen("../ngfile", "w")) == NULL)
		printf("Error: can't open graphpack output file\n"),exit(54);

	fprintf(gpfile, "VERTICES:{%d",gp(nongreedy_vertices[0].v));
	for (i=1; i<strongly_ng; i++)
		fprintf(gpfile, ",%d", gp(nongreedy_vertices[i].v));
	fprintf(gpfile, "}\n");

	fprintf(gpfile, "VWEIGHTS:*%d",gp(nongreedy_vertices[0].v));
	for (i=1; i<strongly_ng; i++)
		fprintf(gpfile, ",%d", gp(nongreedy_vertices[i].v));
	fprintf(gpfile, "*\n");

	fprintf(gpfile, "EDGES:{");
	start = FALSE;
	for (i=0; i<strongly_ng; i++) {
	     	for (j=i+1; j<strongly_ng; j++) {
		    if (adj(nongreedy_vertices[i].v, nongreedy_vertices[j].v)) {
			edges+=1;
			if (!start) {
				fprintf(gpfile,"{%d,%d}",
					gp(nongreedy_vertices[i].v),
					gp(nongreedy_vertices[j].v));
				start = TRUE;
			} else {
				fprintf(gpfile,", {%d,%d}",
					gp(nongreedy_vertices[i].v),
					gp(nongreedy_vertices[j].v));
			}
			fprintf(gpfile,", {%d,%d}",
				gp(nongreedy_vertices[j].v),
				gp(nongreedy_vertices[i].v));
		    }
		}
	}
	fprintf(gpfile,"}\n");

	fprintf(gpfile,"EWEIGHTS:*");
	for(i=0; i<edges-1; i++)
		fprintf(gpfile, ",");
	fprintf(gpfile,"*\n");

	fclose(gpfile);
	fclose(gpinfo);
}
