/***********************************************************/
/*                  Graph Paritioning                      */
/*                        read.c                           */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This module provides the routines to read and       */
/*  validate input data for the graph partitioning         */
/*  routines.  Both common data and data specific to the   */
/*  partitioning method being used are inputed and passed  */
/*  along to the calling program.                          */
/***********************************************************/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "part.h"
#include "data_sets.h"

int scan_int(int, int);
double scan_double(double, double);
char * scan_string();



extern int output_short;
extern int Heuristic_Only;	/* skip postproc,checks, use cpu_seconds   */
extern int Near_Greedy;		/* if TRUE, chk that num=(DISTRIB_SIZE*k)  */
extern int N_Interval;		/* num / NG_UP_SIZE, the ng_uptab size     */
extern int P_Interval;		/* num / DISTRIB_SIZE, the # v in 1% of all*/
extern int N_Prob;		/* NG_Up_Results is discrete, divided      */
				/* into rows of different edge prob. find  */
				/* out which of these is closest to prob   */
extern int P_Prob;		/* ng_distrib is discrete, divided         */
				/* into rows of different edge prob. find  */
				/* out which of these is closest to prob   */
extern int Geometric_Graphs;	/* create these instead of (n,p) ?	   */
extern int Regular_Graphs;	/* create these instead of (n,p) ?	   */
extern int Unbalanced_Regular_Graphs;
extern int Unbalanced_Random_Graphs;
extern int VG_Flag;
extern int PG_VG_Flag;
extern int PG_VG_Flag2;
extern int Find_Clusters;

							/*****WILL CHANGE****/
extern STATS_ARRAY_PTR Var_Stats;			/*****TO BE MORE ****/
							/*****GENERAL    ****/
/*#ifndef KL*/
STATS_ARRAY_PTR get_stats_geom(int, double);
STATS_ARRAY_PTR get_stats_rand(int, double);
/*#endif*/

extern PG_DIST ng_distrib_rand[][DISTRIB_SIZE];	/** ng dist tables for pg **/
extern PG_DIST ng_distrib_geom[][DISTRIB_SIZE];
extern PG_DIST (*ng_distrib)[DISTRIB_SIZE];	/** points to one of these**/

extern PCT_NG_UP up_distrib_rand[][DISTRIB_SIZE];/** ng up dist (k/total) **/
extern PCT_NG_UP up_distrib_geom[][DISTRIB_SIZE];
extern PCT_NG_UP (*NG_Up_Results)[DISTRIB_SIZE]; /** points to one of these*/

extern FILE *outfile, *infile, *efile ;
extern char efilename[20];
extern double min_mult;

/*****************  input data  ******************************************/
int num ;              /** number of vertices 				**/
double prob;            /** edge probability (also dist for geom graphs)**/
double prob2;           /** edge probability 2 for unbal. rand graphs   **/
long seed ;            /** seed 					**/
int num_passes;        /** number of iterations (if not Heuristic_Only) **/
double cpu_seconds;     /** time to run algorithm. 			**/
int sort_type;         /** post processing sort type 			**/
int concurrent_even;   /** concurrent even selection 			**/
int conc_tie ;         /** concurrent find greedy vertex if tie 	**/
int look_ahead;        /** look at unplaced vertices when computing a-b **/
int lmt ;              /** independent set limit 			**/
int expected;          /** independent set expected 			**/
int bounds[MAX_IND_SIZE]; /** independent set bounds 			**/
int num_to_avg;        /** number of vertices to average non-greedy 	**/
/*************************************************************************/

int alg_type;
int read_file_mode ;  /** read from terminal or file? **/

int end_of_file, process; /** process inputed data? **/

int  Exhaustive_Search;		/** tells bk to search exhaustively for opt **/
int  NG_Max;			/** search exh., but only up to <NG_Max> ng.**/
int  Augment_With_Local;
int  Augment_With_KL;
int  Compare_Partitions;
int  Print_Cand_History;
int  Paths;
int  Reg_Degree;		/** k of k-regular graphs                 **/
int  Reg_Degree2;		/** k2 of unbalanced k1-k2-regular graphs **/
int  Change_Index;		/** change the placement (in post_proc()) **/
				/** of the <Change_Index>th vertex, in    **/
				/** placement order.			  **/

void prompt(char *msg)
{
	if (!read_file_mode)
		printf("%s\n",msg);
}

extern int scan_int();
extern double scan_double();
extern char * scan_string();

/*************************************************/
/*********  read one set of input data  **********/
/*************************************************/
/*--- read input data from file or standard input ---*/
int read_data()
{  
	extern int Nongreedy_Stats;
	extern int Output_Sng_Distfile;
	extern int Simulated_Annealing;

	int nb=0,i;
	char *str;

	prob = prob2 = 0.006;

	prompt("Enter the number of vertices (0 to end):");
	num = scan_int(0,MAX_VERTICES);
	/*fscanf(infile, "%d", &num);*/
	if(num == 0) {
		end_of_file = TRUE ;
		process = FALSE ;
		return(TRUE);
	} else if (Output_Sng_Distfile || Near_Greedy) {
		while (num % DISTRIB_SIZE != 0) {
			prompt("Number must be a multiple of DISTRIB_SIZE:");
			num = scan_int(0,MAX_VERTICES);
		}
		P_Interval = (int)  (num / DISTRIB_SIZE);
		N_Interval = (int)  (num / NG_UP_SIZE);
	}

	if(!efile)  { /** no external edge file **/
		if (Geometric_Graphs) {
			prompt("Enter the distance threshold (0.0001..0.999):");
			prob = scan_double(0.0001,0.999);
			ng_distrib = ng_distrib_geom;
			NG_Up_Results = up_distrib_geom;
/*#ifndef KL*/
			if (PG_VG_Flag||PG_VG_Flag2 || Find_Clusters || VG_Flag)
				Var_Stats = get_stats_geom(num,prob); 
/*#endif*/
		} else if (Regular_Graphs) {
		    do {
			prompt("k-regular graphs, num or k must be even k: ");
			Reg_Degree = scan_int(0,num-1);
		    } while ( ((num * Reg_Degree) % 2) != 0 );
		    ng_distrib = ng_distrib_geom;  /****change!****/
		    NG_Up_Results = up_distrib_geom;
		} else if (Unbalanced_Regular_Graphs) {
		    if (num % 2 != 0) {
			prompt("k1-k2-regular graphs, num must be even. num: ");
			num = scan_int(0,MAX_VERTICES);
		    }
		    prompt("k1-k2-regular graphs,  k1, k2: ");
		    Reg_Degree = scan_int(0,num-1);
		    Reg_Degree2 = scan_int(0,num-1);
		    ng_distrib = ng_distrib_geom;	/***********/
		    NG_Up_Results = up_distrib_geom;
		} else if (Unbalanced_Random_Graphs) {
			prompt("Enter the intra-edge prob. (0.0001..0.999):");
			prob = scan_double(0.0001,0.999);
			prompt("Enter the inter-edge prob. (0.0001..0.999):");
			prob2 = scan_double(0.0001,0.999);
		     	ng_distrib = ng_distrib_rand;	/***********/
		     	NG_Up_Results = up_distrib_rand;
		} else {
			prompt("Enter the edge probability (0.0001..0.999):");
			prob = scan_double(0.0001,0.999);
			ng_distrib = ng_distrib_rand;
			NG_Up_Results = up_distrib_rand;
/*#ifndef KL*/
			if (PG_VG_Flag||PG_VG_Flag2||Find_Clusters||VG_Flag)
				Var_Stats = get_stats_rand(num,prob); 
/*#endif*/
		}

		/******************************************************/
		/* Determine which row of ng_distrib[][] to use      */
		/******************************************************/
		if (Near_Greedy) {
		   if (Geometric_Graphs) {
			P_Prob = 0;				/** 12/22/93 **/
			/*if (prob <= 0.125)	P_Prob = 0;
			else 			P_Prob = 1;*/
			/********** temp changes 11/92 ***********/
			/*if (prob <= 0.075)	P_Prob = 0;
			else if (prob <= 0.125)	P_Prob = 1;
			else if (prob <= 0.175)	P_Prob = 2;
			else if (prob <= 0.225) P_Prob = 3;
			else 			P_Prob = 4;*/
			/********** temp changes 11/92 ***********/
		   } else if (Regular_Graphs) {
			prob = (double) sqrt( (double)Reg_Degree / (num * M_PI));
			if (prob <= 0.125)	P_Prob = 0;
			else 			P_Prob = 1;
			/********** temp changes 11/92 ***********/
			/*if (prob <= 0.15)	P_Prob = 0;
			else if (prob <= 0.25)	P_Prob = 1;
			else if (prob <= 0.35)	P_Prob = 2;
			else if (prob <= 0.45)	P_Prob = 3;
			else 			P_Prob = 4;*/
			/********** temp changes 11/92 ***********/
		   } else {
			if (prob <= 0.125)	P_Prob = 0;
			else 			P_Prob = 1;
			/********** temp changes 11/92 ***********/
			/*if (prob <= 0.15)	P_Prob = 0;
			else if (prob <= 0.25)	P_Prob = 1;
			else if (prob <= 0.35)	P_Prob = 2;
			else if (prob <= 0.45)	P_Prob = 3;
			else 			P_Prob = 4;*/
			/********** temp changes 11/92 ***********/
		   }
		}
		/******************************************************/
		/* Determine which row of NG_Up_Results[][] to use      */
		/******************************************************/
		if (Nongreedy_Stats || Near_Greedy) {
		   if (Geometric_Graphs) {
			N_Prob = 0;
			if (prob <= 0.075)	N_Prob = 0;
			else			N_Prob = 1;
			/*else if (prob <= 0.125)	N_Prob = 1;
			else if (prob <= 0.175)	N_Prob = 2;
			else if (prob <= 0.225) N_Prob = 3;
			else 			N_Prob = 4;*/
		   } else if (Regular_Graphs) {
			if (prob <= 0.075)	N_Prob = 0;
			else			N_Prob = 1;
			/*else if (prob <= 0.25)	N_Prob = 1;
			else if (prob <= 0.35)	N_Prob = 2;
			else if (prob <= 0.45)	N_Prob = 3;
			else 			N_Prob = 4;*/
		   } else {
			if (prob <= 0.075)	N_Prob = 0;
			else			N_Prob = 1;
			/*else if (prob <= 0.25)	N_Prob = 1;
			else if (prob <= 0.35)	N_Prob = 2;
			else if (prob <= 0.45)	N_Prob = 3;
			else 			N_Prob = 4;*/
		   }
		}
		/*****************************************************/
     
		prompt("Enter the random number seed:");
		seed = scan_int(0,MAX_INT);
	}
     
	if (alg_type == ALG_TYPE_KL) {
		if (!Heuristic_Only) {
			prompt("Enter number of iterations:");
			num_passes = scan_int(0, MAX_PASSES);  
		} else {
			prompt("Enter time limit in CPU seconds:");
			cpu_seconds = scan_double(0.0, MAX_TIME);
		}
	}

	if (alg_type==ALG_TYPE_KL) {
		prompt("Enter the Post-Processing sort type:");
		if (!output_short) {
			fprintf(outfile,"   0 - random\n");
			fprintf(outfile,"   1 - decreasing degree\n");
			fprintf(outfile,"   2 - increasing degree\n");
			fprintf(outfile,"   3 - forward decreasing\n");
			fprintf(outfile,"   4 - forward increasing\n");
			fprintf(outfile,"   5 - single independent set\n");
			fprintf(outfile,"   6 - multiple independent sets\n");
			fprintf(outfile,"   7 - concurrent reordering\n");
			fprintf(outfile,"   8 - ind. set with conc. reord.\n");
			fprintf(outfile,"   9 - backtracking\n");
			fprintf(outfile,"   10 -subgraph max. deg.\n");
		}
		sort_type = scan_int(0,10);
		sort_type = (sort_type == 9) ? BACKTRACK : sort_type;
	} 

   if(sort_type == 5 || sort_type == 6
         || sort_type == 8) {  /** input independent set parms **/
      prompt("Enter the expected indpendent set size:");
      expected = scan_int(1,MAX_IND_SIZE);

      prompt("Enter the limit of the bounds:");
      lmt = scan_int(1,MAX_INT);

      prompt("Enter the number of bounds you wish to enter:") ;
      prompt("NOTE: the last one entered will be duplicated to the end.");
      nb = scan_int(1,MAX_INT);

      prompt("Enter each bound separated by a space:");
      for(i=0;i<nb;i++) 
	 bounds[i] = scan_int(1,MAX_INT);
      for(i=nb;i<MAX_IND_SIZE;i++)
         bounds[i] = bounds[nb-1];
    }

   if((sort_type == 7 || sort_type == 8) && (alg_type != BACKTRACKING)) { 
      /** input concurrent reordering parameters **/
	prompt("With concurrent reordering what should be done when there");
	prompt("  are ties: 1-take highest degree 2- take lowest degree");
	concurrent_even = scan_int(1,2);

      /** input concurrent reordering on tie search for greedy **/
      conc_tie = 0 ;
      if(alg_type == ALG_TYPE_KL) {
         prompt("If tie, attempt to find greedy vertex: 1-Yes 0-No");
	 conc_tie = scan_int(0,1);
      }
       
    }

   if(sort_type != BACKTRACK && alg_type != BACKTRACKING) {
      prompt("Do you wish to look at unplaced vertices when computing");
      prompt("   greedy/non-greedy steps: 0-No 1-Yes");
      look_ahead = scan_int(0,1);
   }

   prompt("Change the placement of the i'th vertex.  i = ?"); 
   Change_Index = scan_int(0, num+1) - 1;

   if(alg_type == BACKTRACKING) {
	prompt("Search exhaustively for optimal? (1/0):");
	Exhaustive_Search = scan_int(0,1);
	if (Exhaustive_Search) {
		prompt("Enter max. proportion nongreedy (.0001-.999):");
		NG_Max = (int) (num * scan_double(0.0001, 0.999));
		Paths = BACKTRACK;
	} else {
		if (!Heuristic_Only) {
			prompt("Enter the number of random paths to follow");
			prompt("(0:backtrack, (x>0):follow x random paths)");
			Paths = scan_int(0, MAX_INT);
			Paths = (Paths == 0) ? BACKTRACK : Paths;
		} else {
			prompt("Enter the time to run in CPU seconds");
			prompt("(0:backtrack, (x>0):run for x seconds)");
			cpu_seconds = scan_double(0.0, MAX_TIME);
		}
	
		prompt("Augment with Local Optimization? (0:no, 1:yes)");
		Augment_With_Local = scan_int(0, 1);

		if (!Augment_With_Local) {
			prompt("Augment with KL? (0:no, 1:yes)");
			Augment_With_KL = scan_int(0, 1);
		}
	
		prompt("Run entire KL and Compare partitions? (0:no, 1:yes)");
		Compare_Partitions = scan_int(0, 1);
	
		prompt("Print Candidate Histories? (0:no, 1:yes)");
		Print_Cand_History = scan_int(0, 1);
	}
   }

   prompt("Is this information correct? (y to processs):");
   str = scan_string();
   if(str[0] != 'y' && str[0] != 'Y') {
      if((strcmp(str,"quit") == 0)||
	 (strcmp(str,"stop") == 0)||(strcmp(str,"exit") == 0))
         end_of_file = TRUE ;
      process = FALSE ;
      free(str);
      return(TRUE);
    }

   /*print_summary();*/

	/*******************************************************************/
	/***  print report of input data  **********************************/
	/*******************************************************************/
	if (Simulated_Annealing)
		fprintf(outfile,"\n\n*****  Simulated Annealing  *****\n");
	else if (alg_type==ALG_TYPE_KL)
		fprintf(outfile,"\n\n****  Kernighan Lin  (FM)  ****\n");
	else if (alg_type == BACKTRACKING)
		fprintf(outfile,"\n\n*****  Backtracking *****\n");
   	fprintf(outfile,"-----------------------------------\n");
   	fprintf(outfile,"INPUT DATA:\n");
   	fprintf(outfile,"   number of vertices: %d\n",num);
	if(efile)   /** external edge probability **/
		fprintf(outfile,"   edge probability: %s\n",efilename);
	else {
		if (Geometric_Graphs) 
			fprintf(outfile,"   distance threshold: %.6f\n",prob);
		else if (Regular_Graphs) 
			fprintf(outfile,"   k-regular, k = : %d\n",Reg_Degree);
		else
			fprintf(outfile,"   edge probability: %.6f\n",prob);
		fprintf(outfile,"   seed: %ld\n",seed);
	}
	if (!Heuristic_Only)
		fprintf(outfile," passes: %d\n",num_passes);
	else
		fprintf(outfile," time to run: %.6f\n",cpu_seconds);
	fprintf(outfile,"   sort: ");
	switch(sort_type) {
		case 0 : fprintf(outfile,"random\n");
			break;
		case 1 : fprintf(outfile,"decreasing degree\n");
			break;
		case 2 : fprintf(outfile,"increasing degree\n");
			break;
		case 3 : fprintf(outfile,"forward decreasing degree\n");
			break;
		case 4 : fprintf(outfile,"forward increasing degree\n");
			break;
		case 5 :
		case 6 : 
		case 8 :
			if(sort_type == 5)
				fprintf(outfile,"single independent set\n");
			else if(sort_type == 6)
				fprintf(outfile,"multiple indpendent set\n");
			else
				fprintf(outfile,"indep. set with conc reord\n");
			fprintf(outfile,"         expected: %d\n",expected);
			fprintf(outfile,"            limit: %d\n",lmt);
			fprintf(outfile,"           bounds: ");
			for(i=0;i<nb;i++)
				fprintf(outfile,"%d ",bounds[i]);
			fprintf(outfile,"...\n");
			if(sort_type != 8) /** ind w/ conc. reordering **/
				break;
		case 7 :fprintf(outfile,"concurrent reordering\n");
			fprintf(outfile,"  tie selection: ");
			if(concurrent_even == 1)
				fprintf(outfile,"highest degree\n");
			else if(concurrent_even == 2)
				fprintf(outfile,"lowest degree\n");
			else
				fprintf(outfile,"???????????????\n");
			if(alg_type == ALG_TYPE_KL) {
				fprintf(outfile,"tie find greedy: ");
				if(conc_tie)
					fprintf(outfile,"Yes\n");
				else
					fprintf(outfile,"No\n");
			}
			break;
		case BACKTRACK:  fprintf(outfile,"backtracking\n");
			break;
		case 10 : fprintf(outfile,"Subgraph Maxdegree\n");
			break;
		default: fprintf(outfile,"*****unknown*****\n");
	}

	fprintf(outfile,"     look ahead: ");
	if(look_ahead)
		fprintf(outfile,"Yes\n");
	else
		fprintf(outfile,"No\n");

	min_mult = MULTIPLIER ;
	process = TRUE ;
	free(str);
	return(TRUE);
}
