#include<stdio.h>
/*Hi Jon,

Here are a few large graphs and the corresponding results. Upper_bd
refers to the positive semidefinite program. The lower bound is
what my implementation of the GW algorithm finds within a few
minutes. Let me know if you find anything larger. Which algorithm
are you using?

Sparse random graphs G(n, 10/n)

file	n	edges	upper_bd	lower_bd
=================================================
r7000.1	6999	35319	27865		24964
r8000.1	7999	39824	31510		28255
r8000	8000	39951	31585		28307
r7000	6999	34927	27600		24725
r6000	6000	30138	23770		21311
r5000	5000	24938	19690		17671



Sparse geometric graphs G(d,d,0.5,2)

file	d	n	edges	upper_bd	lower_bd
========================================================
g100	100	5029	14805	10965		10356
g100.1	100	4927	14405	10680		10076



About the graph format: The graphs are stored as edge lists, where
each edge has the following form:
*/
/*#define MAX_NR_EDGES 80000

typedef struct {
  int v1, v2;
  double wt;
} F_EDGE;*/      /* the form in which the edges are stored in the file */

#include "read_graph.h"

/*int fread (char*, int, int, FILE*);*/
/*int    printf(const char*, ...);*/
/*int    fclose(FILE *);*/


F_EDGE edge_list[MAX_NR_EDGES];
int nr_dir_edges;

void read_graph( char *filename, int exp_nr_edges )
{
  FILE * fp;

  fp = fopen( filename,"r" );
  if ( fp == NULL ) exit(31);

  nr_dir_edges = fread( (void *)edge_list, sizeof( F_EDGE ), MAX_NR_EDGES, fp );
  if (nr_dir_edges != (2*exp_nr_edges))
	printf("Error: number of edges different from expectation\n");

  fclose( fp );
}

void read_peinado(char *filename)
{
	if ((strcmp(filename, "r7000.1")) == 0)
		read_graph("r7000.1", 35319);
	else if ((strcmp(filename, "r8000.1")) == 0)
		read_graph("r8000.1", 39824);
	else if ((strcmp(filename, "r8000")) == 0)
		read_graph("r8000", 39951);
	else if ((strcmp(filename, "r7000")) == 0)
		read_graph("r7000", 34927);
	else if ((strcmp(filename, "r6000")) == 0)
		read_graph("r6000", 30138);
	else if ((strcmp(filename, "r5000")) == 0)
		read_graph("r5000", 24938);
	else if ((strcmp(filename, "g100")) == 0)
		read_graph("g100", 14805);
	else if ((strcmp(filename, "g100.1")) == 0)
		read_graph("g100.1", 14405);
	else 
	       printf("Error: illegal Peinado graph specified: %s\n", filename);
	printf("edges: %d\n", nr_dir_edges);
}

/*One thing to keep in mind is that the graph is stored as a directed graph --
i.e. every edge appears twice: as (v1,v2,wt) and as (v2,v1,wt).*/
