/***********************************************************/
/*                  Graph Paritioning                      */
/*                        sort.c                           */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This module provides the routines to reorder the    */
/*  graph vertex by the specified criteria.  Sorted vertex */
/*  numbers are placed in the "vertices" array to be used  */
/*  my the calling program.                                */
/***********************************************************/
/**  Valid reordering types:                              **/
/**       (0) random                                      **/
/**       (1) decreasing degree                           **/
/**       (2) increasing degree                           **/
/**       (3) forward decreasing degree                   **/
/**       (4) forward increasing degree                   **/
/**       (5) single independent set                      **/
/**       (6) multiple independent sets                   **/
/**       (7) concurrent reordering                       **/
/**       (8) independent set with concurrent reordering  **/
/***********************************************************/

#include <stdio.h>
#include <ctype.h>

#include "part.h"

extern int deg[] ;
extern int vertices[] ;
extern int num ;
extern int cum_even ;

extern struct adj_node *adj_list[] ;

extern int sort_type, alg_type;
extern int concurrent_even;
extern FILE *outfile;

int last ;

/*** independent set variables ***/
int max_size ;
int line[MAX_VERTS];
int length[MAX_VERTS];
int max_set[MAX_VERTS];
int start[MAX_VERTS];
int full[MAX_VERTS];
int finish[MAX_VERTS];
int last_bounds[MAX_VERTS];
int chislo[MAX_VERTS];
int indx, prdct ;
int iassigned[MAX_VERTS];
int inum,iplace;

int is_ind_set[MAX_VERTS];

extern int lmt, expected;
extern int bounds[];

/*********************************************************/
/*****  compare 2 integers using reordering choice  ******/
/*********************************************************/
int compare(int p1,int p2)
{
   if(sort_type == 1 || sort_type ==3 || sort_type == 9) /** decreasing **/
      if(p1>p2)
	 return(TRUE);
      else
	 return(FALSE);
   else  /** increasing **/
      if(p1<p2)
	 return(TRUE);
      else
	 return(FALSE);
   }

/*********************************************************/
/*********  check if two vertices are adjacent   *********/
/*********************************************************/
int is_adjacent(int a,int b)
{  struct adj_node *p;
   int f ;
   if(a > b) {
     p = adj_list[b];
     f = a ;
   }
   else {
     p = adj_list[a] ;
     f = b ;
   }
   while(p != NULL) {
      if(p->node_no==f)
         return(TRUE);
      if(p->node_no <f)
         return(FALSE);
      p = p->next ;
    }
   return(FALSE);
 }
      
/*****************************************************/
/*****  initialize independent set variables   *******/
/*****************************************************/
void init_ind_set()
{  int i;
   max_size = 0 ;
   indx = 0 ;
   for(i=0;i<MAX_IND_SIZE;i++) {
      chislo[i] = -1 ;
      bounds[i]--;
      max_set[i] = -1 ;
      finish[i] = -1 ;
      full[i] = -1 ;
      start[i] = -1 ;
      last_bounds[i] = -1 ;
      length[i] = 0 ;
    }
   length[0] = inum;
   start[0] = 0 ;
   full[0] = inum -1 ;
   if(bounds[0] <inum-1)
      finish[0] = bounds[0] ;
   else
      finish[0] = inum -1 ;
   if(finish[0] > lmt)
      finish[0] = lmt ;
   prdct = 0 ;
 }

/**********************************************************************/
/*****  place vertices not assigned in line for ind. set routines  ****/
/**********************************************************************/
int reorder_ind_set()
{  int i;
   inum = 0 ;
   for(i=0;i<num;i++)
      if(!iassigned[i]) 
         line[inum++] = i ;
   return(inum);
 }

/********************************************************/
/*****  vertify that solution is an indpendent set  *****/
/********************************************************/
void check_solution()
{  int i,j, indep ;
   indep = TRUE ;
   for(i=0;i<max_size-1;i++)
      for(j=i+1;j<max_size;j++)
         if(is_adjacent(max_set[i],max_set[j]))
            indep = FALSE ;
   if(!indep) {
      printf("error in independent set routine\07\07\07\n");
      exit(909);
    }
 }


/**********************************************************/
/*    This routine finds a maximal independent set to be  */
/* placed at the end of the list of vertices.             */
/**********************************************************/
/*    Original code for this routine was provided by Mark */
/*  Goldberg on May 24, 1990.  Minor adjustments were     */
/*  made.                                                 */
/**********************************************************/
void find_ind_set()

/***** used to have: str[MAX_IND_SIZE][MAX_VERTICES] - got too big jwb 1/94 ***/
{ int i, yy, go, sm, prd, cnt=0, str[2][1000];

   for(i=0; i<inum; i++)
      str[0][i] = line[i];
   go = 1;
   while (indx >= 0) {
      if (go == 1) {
         yy = str[indx][start[indx]];
	 cnt = 0;
	 for (i = start[indx]+1; i <= full[indx]; i++) {
            if (!is_adjacent(yy,str[indx][i])) {
               str[indx+1][cnt] = str[indx][i];
	       cnt++;
	     }
	  }
	 if( cnt > length[indx + 1])
	    length[indx + 1] = cnt;
       }
      if ( (cnt > 0) && (indx + cnt >= max_size) )  {  /*FORWARD */
	 indx++;
	 full[indx]=cnt-1;
	 start[indx]=0;
	 if ( full[indx] > bounds[indx] )
	    finish[indx] = bounds[indx];
	 else
	    finish[indx] = full[indx];
       }
      else {
         if (indx >= max_size) {  /*UPDATE */
	    max_size = indx+1;
	    sm = start[0];
	    prd = start[0] + 1;
	    for (i=1; i < max_size; i++) {
	       sm = sm + start[i];
	       prd = prd * (start[i] +1);
	     }
	    chislo[max_size] = sm;
	    for (i=0; i < max_size; i++) {
               last_bounds[i] = start[i];
	       max_set[i] = str[i][start[i]];
	     }
	    if (max_size >= expected)
	       break;
	  }

         if  ((start[indx] < finish[indx]) && (prdct < lmt )) { /*LEFT*/
	    start[indx]++;
	    prdct = prdct++;
	    go = 1;
          }
         else { /* BACK */
	    prdct = prdct-start[indx];
	    indx--;
	    go = 0;
          }
       }
    }
}

/******************************************************************/
/****  add the degrees of a placed vertex for independent set  ****/
/******************************************************************/
void add_ideg(int v,int *ideg)
{  struct adj_node *p;
   p = adj_list[v];
   while(p != NULL) {
      if(ideg[p->node_no] != SKIP_VERTEX)
         ideg[p->node_no]++;
      p = p->next ;
    }
   ideg[v] = SKIP_VERTEX;
 }

/*******************************************************************/
/***  reorder by finding a single independent set and placing    ***/
/***     at the end of the list of vertices.  remaining vertices ***/
/***     are placed starting at the independent set and going    ***/
/***     backwards from lowest to highest degree into the already **/
/***     placed vertices.                                        ***/
/*******************************************************************/
void single_ind_set_order()
{  int i, j, lowest_deg, lowest_ideg, holdi;
   int ideg[MAX_VERTS];
   for(i=0;i<num;i++)
      ideg[i] = 0 ;
   i = num-1;
   for(j=0;j<max_size;j++) {
      vertices[i] = max_set[j] ;
      is_ind_set[vertices[i]] = TRUE;
      add_ideg(vertices[i],ideg);
      i--;
    }
   for(j=i;j>=0;j--) {
      lowest_deg = lowest_ideg = VERY_HIGH ;
      holdi = 0 ;
      for(i=0;i<num;i++) {
         if(ideg[i] != SKIP_VERTEX)
            if(ideg[i] < lowest_ideg ||
                    (ideg[i] == lowest_ideg && deg[i] < lowest_deg)) {
               lowest_ideg = ideg[i] ;
               lowest_deg = deg[i] ;
               holdi = i ;
	     }
       }
      vertices[j] = holdi;
      add_ideg(holdi,ideg);
    }
 }

/***************************************************************/
/**  place found independent set in the next open places from **/
/**      the end of already placed vertices.                  **/
/***************************************************************/
void mult_ind_set_order()
{  int i;
   for(i=0;i<max_size;i++) { 
      vertices[--iplace] = max_set[i];
      iassigned[max_set[i]] = TRUE ;
    }
 }

/************************************************************/
/****  reorder by finding a single independent set.      ****/
/************************************************************/
void single_independent_set()
{  int i;
   for(i=0;i<num;i++)
      iassigned[i] = FALSE ;
   inum = num;
   init_ind_set();
   for(i=0;i<num;i++)
      line[i] = i;
   find_ind_set();
   check_solution();
   fprintf(outfile,"independent set of size %d found.\n",max_size) ;
   single_ind_set_order();
 }

/*******************************************************/
/***  reorder by finding multiple independent sets   ***/
/*******************************************************/
void mult_independent_set()
{  int i;
   iplace = num;
   for(i=0;i<num;i++)
      iassigned[i] = FALSE ;
   while(reorder_ind_set() > 0 ) {
      init_ind_set();
      find_ind_set();
      check_solution();
      fprintf(outfile,"independent set of size %d found.\n",max_size) ;
      mult_ind_set_order();
    }
   if(iplace != 0)
      printf("internal error #1011 in sort.c\07\07\07\n");
}


/***************************************************/
/***********  remove edges from vertex  ************/
/***************************************************/
void remove_vertex(int v)
{  struct adj_node *p ;
   p = adj_list[v] ;
   while(p != NULL) {
      if(deg[p->node_no] != SKIP_VERTEX)
         deg[p->node_no]--;
      p = p->next ;
    }
  }
 
/*****************************************************/
/********  place vertices into sorted order  *********/
/*****************************************************/
int get_order_vertex() {  /** get next vertex in sorted list **/
   register i;
   int highlow, phighlow=0;
   if(sort_type == 1 || sort_type == 3 || sort_type == 9)
      highlow = VERY_LOW;
   else
      highlow = VERY_HIGH;
   /*** this is done because in deleting of vertices, we may have vertices
	     with one lower number of edges than the previous one ***/
   if(sort_type == 4) /** forward increasing **/
      last = last -1;

   for(i=0;i<num;i++) {
      if(deg[i] != SKIP_VERTEX) {
	 if(deg[i] == last) {
	    deg[i] = SKIP_VERTEX;
	    if(sort_type == 3 || sort_type == 4)
	       remove_vertex(i);  /** for. inc./for. dec. **/
	    return(i);
	  }
	 if(compare(deg[i],highlow)) {
	    highlow = deg[i];
	    phighlow = i;
	  }
       }
    }
   last = deg[phighlow];
   deg[phighlow] = SKIP_VERTEX;
   if(sort_type == 3 || sort_type == 4)  /** for. inc./for. dec. **/
      remove_vertex(phighlow);
   return(phighlow);
 }

/***************************************************************/
/**  should vertices of equal degree difference be switched.  **/
/***************************************************************/
int better_degree(int curi,int highesti,int first)
{  
   if(concurrent_even == 2 && !first) /** on even take lowest degree **/
      if(deg[curi] < deg[highesti])
         return(TRUE);
      else
         return(FALSE);
   else  /** on even take highest degree **/
      if(deg[curi] > deg[highesti])
         return(TRUE);
      else
         return(FALSE);
 }

/*****************************************************************/
/***  Reorder using concurrent reordering, which says to place ***/
/***    vertices of the largest "greedy" degree based on the   ***/
/***    vertices which have already been placed.               ***/
/*****************************************************************/
void concurrent_reordering()
{  int vertex_loc[MAX_VERTS],node_cost[2][MAX_VERTS];
   int highestj,highest,val;
   register i,j;
   struct adj_node *p;
   for(i=0;i<num;i++) {
      vertex_loc[i] = SKIP_VERTEX;
      node_cost[SET_A][i] = 0 ;
      node_cost[SET_B][i] = 0 ;
    }
   for(i=0;i<num;i++) {
      highestj = -1;
      highest = VERY_LOW ;
      for(j=0;j<num;j++) {
         if(vertex_loc[j] == SKIP_VERTEX) {
            val = ABS(node_cost[SET_A][j] - node_cost[SET_B][j]);
            if(val > highest || 
              (val == highest && better_degree(j,highestj,!i))) {
               highestj = j;
               highest = val;
               }
          }
       }
      if(node_cost[SET_A][highestj] >= node_cost[SET_B][highestj]) 
         vertex_loc[highestj] = SET_A;
      else
         vertex_loc[highestj] = SET_B;
      vertices[i] = highestj;
      p = adj_list[highestj];
      while(p != NULL) {
         if(vertex_loc[highestj])
            node_cost[SET_A][p->node_no]++;
         else
            node_cost[SET_B][p->node_no]++;
         p = p->next ;
       }
       
    }
 }

/******************************************************/
/************  preform vertice reordering  ************/
/******************************************************/
void process_reorder()  {
   register i;
   /*int hdeg[MAX_VERTS];*/
/*
   for(i=0;i<num;i++) {
      is_ind_set[i] = FALSE;
      hdeg[i] = deg[i];
    }
*/
   switch(sort_type) {
   case 0:  for(i=0;i<num;i++) /** random **/
               vertices[i] = i ;
            break;
   case 1:  /** decreasing degree **/
   case 2:  /** increasing degree **/
   case 3:  /** forward decreasing degree **/
   case 4:  /** forward increasing degree **/
            last = SKIP_VERTEX ;
            i = 0 ;
            while(i<num)
               vertices[i++] = get_order_vertex();
            break;
   case 5 : single_independent_set();
            break;
   case 6:  mult_independent_set();
            break;
   case 7:  if(alg_type == ALG_TYPE_NEAR_GREEDY) 
               concurrent_reordering();
            else /** concurrent reordering with K-L must be done concurrently
                        with post-processing **/
               for(i=0;i<num;i++) /** done in post-processing **/
                  vertices[i] = i ;   
            break;
   case 8:  single_independent_set(); /** ind set + conc. reordering **/
            break; 
   case 10: for(i=0;i<num;i++) /** done in post-processing **/
                  vertices[i] = i ;   
            break;
   case BACKTRACK:  
            last = SKIP_VERTEX ;
            vertices[0] = get_order_vertex();
            break;
   default: printf("invalid sort type in sort.c\07\07\07\n");
   }
   /*for(i=0;i<num;i++)
      deg[i] = hdeg[i];*/
   if(num < 10) {   /**** test mode  ****/
      printf("sorted vertex order:");
      for(i=0;i<num;i++)
	 printf(":%d:",vertices[i]);
      printf("\n");
    }
 }


